from django.template import Context, Template
from django.test import TestCase
from jinja2 import Environment, PackageLoader

from pipeline.jinja2 import PipelineExtension
from tests.utils import pipeline_settings


class JinjaTest(TestCase):
    def setUp(self):
        self.env = Environment(
            extensions=[PipelineExtension],
            loader=PackageLoader("pipeline", "templates"),
        )

    def test_no_package(self):
        template = self.env.from_string("""{% stylesheet "unknow" %}""")
        self.assertEqual("", template.render())
        template = self.env.from_string("""{% javascript "unknow" %}""")
        self.assertEqual("", template.render())

    def test_package_css(self):
        template = self.env.from_string("""{% stylesheet "screen" %}""")
        self.assertEqual(
            '<link href="/static/screen.css" rel="stylesheet" type="text/css" />',
            template.render(),
        )

    @pipeline_settings(PIPELINE_ENABLED=False)
    def test_package_css_disabled(self):
        template = self.env.from_string("""{% stylesheet "screen" %}""")
        self.assertEqual(
            """<link href="/static/pipeline/css/first.css" rel="stylesheet" type="text/css" />
<link href="/static/pipeline/css/second.css" rel="stylesheet" type="text/css" />
<link href="/static/pipeline/css/urls.css" rel="stylesheet" type="text/css" />""",  # noqa
            template.render(),
        )

    def test_package_js(self):
        template = self.env.from_string("""{% javascript "scripts" %}""")
        self.assertEqual(
            '<script type="text/javascript" src="/static/scripts.js" charset="utf-8"></script>',  # noqa
            template.render(),
        )

    def test_package_js_async(self):
        template = self.env.from_string("""{% javascript "scripts_async" %}""")
        self.assertEqual(
            '<script async type="text/javascript" src="/static/scripts_async.js" charset="utf-8"></script>',  # noqa
            template.render(),
        )

    def test_package_js_defer(self):
        template = self.env.from_string("""{% javascript "scripts_defer" %}""")
        self.assertEqual(
            '<script defer type="text/javascript" src="/static/scripts_defer.js" charset="utf-8"></script>',  # noqa
            template.render(),
        )

    def test_package_js_async_defer(self):
        template = self.env.from_string("""{% javascript "scripts_async_defer" %}""")
        self.assertEqual(
            '<script async defer type="text/javascript" src="/static/scripts_async_defer.js" charset="utf-8"></script>',  # noqa
            template.render(),
        )


class DjangoTest(TestCase):
    def render_template(self, template):
        return Template(template).render(Context())

    def test_compressed_empty(self):
        rendered = self.render_template(
            """{% load pipeline %}{% stylesheet "unknow" %}""",
        )
        self.assertEqual("", rendered)

    def test_compressed_css(self):
        rendered = self.render_template(
            """{% load pipeline %}{% stylesheet "screen" %}""",
        )
        self.assertEqual(
            '<link href="/static/screen.css" rel="stylesheet" type="text/css" media="all" />',  # noqa
            rendered,
        )

    def test_compressed_css_media(self):
        rendered = self.render_template(
            """{% load pipeline %}{% stylesheet "screen_media" %}""",
        )
        self.assertEqual(
            '<link href="/static/screen_media.css" rel="stylesheet" type="text/css" media="screen and (min-width:500px)" />',  # noqa
            rendered,
        )

    def test_compressed_css_title(self):
        rendered = self.render_template(
            """{% load pipeline %}{% stylesheet "screen_title" %}""",
        )
        self.assertEqual(
            '<link href="/static/screen_title.css" rel="stylesheet" type="text/css" media="all" title="Default Style" />',  # noqa
            rendered,
        )

    def test_compressed_js(self):
        rendered = self.render_template(
            """{% load pipeline %}{% javascript "scripts" %}""",
        )
        self.assertEqual(
            '<script type="text/javascript" src="/static/scripts.js" charset="utf-8"></script>',  # noqa
            rendered,
        )

    def test_compressed_js_async(self):
        rendered = self.render_template(
            """{% load pipeline %}{% javascript "scripts_async" %}""",
        )
        self.assertEqual(
            '<script async type="text/javascript" src="/static/scripts_async.js" charset="utf-8"></script>',  # noqa
            rendered,
        )

    def test_compressed_js_defer(self):
        rendered = self.render_template(
            """{% load pipeline %}{% javascript "scripts_defer" %}""",
        )
        self.assertEqual(
            '<script defer type="text/javascript" src="/static/scripts_defer.js" charset="utf-8"></script>',  # noqa
            rendered,
        )

    def test_compressed_js_async_defer(self):
        rendered = self.render_template(
            """{% load pipeline %}{% javascript "scripts_async_defer" %}""",
        )
        self.assertEqual(
            '<script async defer type="text/javascript" src="/static/scripts_async_defer.js" charset="utf-8"></script>',  # noqa
            rendered,
        )



from os.path import dirname

from pipeline.compilers import SubProcessCompiler
from pipeline.conf import settings


class SASSCompiler(SubProcessCompiler):
    output_extension = "css"

    def match_file(self, filename):
        return filename.endswith((".scss", ".sass"))

    def compile_file(self, infile, outfile, outdated=False, force=False):
        command = (settings.SASS_BINARY, settings.SASS_ARGUMENTS, infile, outfile)
        return self.execute_command(command, cwd=dirname(infile))



from unittest.mock import patch

from django.core.exceptions import MiddlewareNotUsed
from django.http import HttpRequest, HttpResponse
from django.test import TestCase

from pipeline.middleware import MinifyHTMLMiddleware


def dummy_get_response(request):
    return None


class MiddlewareTest(TestCase):
    whitespace = b"    "

    def setUp(self):
        self.req = HttpRequest()
        self.req.META = {
            "SERVER_NAME": "testserver",
            "SERVER_PORT": 80,
        }
        self.req.path = self.req.path_info = "/"
        self.resp = HttpResponse()
        self.resp.status_code = 200
        self.resp.content = self.whitespace

    def test_middleware_html(self):
        self.resp["Content-Type"] = "text/html; charset=UTF-8"

        response = MinifyHTMLMiddleware(dummy_get_response).process_response(
            self.req, self.resp
        )
        self.assertIn("text/html", response["Content-Type"])
        self.assertNotIn(self.whitespace, response.content)

    def test_middleware_text(self):
        self.resp["Content-Type"] = "text/plain; charset=UTF-8"

        response = MinifyHTMLMiddleware(dummy_get_response).process_response(
            self.req, self.resp
        )
        self.assertIn("text/plain", response["Content-Type"])
        self.assertIn(self.whitespace, response.content)

    @patch("pipeline.middleware.settings.PIPELINE_ENABLED", False)
    def test_middleware_not_used(self):
        self.assertRaises(MiddlewareNotUsed, MinifyHTMLMiddleware, dummy_get_response)



import os

import django
from django.test import override_settings


def _(path):
    # Make sure the path contains only the correct separator
    return path.replace("/", os.sep).replace("\\", os.sep)


class pipeline_settings(override_settings):
    def __init__(self, **kwargs):
        if django.VERSION[:2] >= (1, 10):
            # Django 1.10's override_settings inherits from TestContextDecorator
            # and its __init__ method calls its superclass' __init__ method too,
            # so we must do the same.
            super().__init__()
        self.options = {"PIPELINE": kwargs}



import os
import shutil

from django.core.files.base import ContentFile
from django.core.files.storage import FileSystemStorage
from django.test import TestCase

from pipeline import glob


def local_path(path):
    return os.path.join(os.path.dirname(__file__), path)


class GlobTest(TestCase):
    def normpath(self, *parts):
        return os.path.normpath(os.path.join(*parts))

    def mktemp(self, *parts):
        filename = self.normpath(*parts)
        base, file = os.path.split(filename)
        base = os.path.join(self.storage.location, base)
        if not os.path.exists(base):
            os.makedirs(base)
        self.storage.save(filename, ContentFile(""))

    def assertSequenceEqual(self, l1, l2):
        self.assertEqual(set(l1), set(l2))

    def setUp(self):
        self.storage = FileSystemStorage(local_path("glob_dir"))
        self.old_storage = glob.staticfiles_storage
        glob.staticfiles_storage = self.storage
        self.mktemp("a", "D")
        self.mktemp("aab", "F")
        self.mktemp("aaa", "zzzF")
        self.mktemp("ZZZ")
        self.mktemp("a", "bcd", "EF")
        self.mktemp("a", "bcd", "efg", "ha")

    def glob(self, *parts):
        if len(parts) == 1:
            pattern = parts[0]
        else:
            pattern = os.path.join(*parts)
        return glob.glob(pattern)

    def tearDown(self):
        shutil.rmtree(self.storage.location)
        glob.staticfiles_storage = self.old_storage

    def test_glob_literal(self):
        self.assertSequenceEqual(self.glob("a"), [self.normpath("a")])
        self.assertSequenceEqual(self.glob("a", "D"), [self.normpath("a", "D")])
        self.assertSequenceEqual(self.glob("aab"), [self.normpath("aab")])

    def test_glob_one_directory(self):
        self.assertSequenceEqual(
            self.glob("a*"), map(self.normpath, ["a", "aab", "aaa"])
        )
        self.assertSequenceEqual(self.glob("*a"), map(self.normpath, ["a", "aaa"]))
        self.assertSequenceEqual(self.glob("aa?"), map(self.normpath, ["aaa", "aab"]))
        self.assertSequenceEqual(
            self.glob("aa[ab]"), map(self.normpath, ["aaa", "aab"])
        )
        self.assertSequenceEqual(self.glob("*q"), [])

    def test_glob_nested_directory(self):
        if os.path.normcase("abCD") == "abCD":
            # case-sensitive filesystem
            self.assertSequenceEqual(
                self.glob("a", "bcd", "E*"), [self.normpath("a", "bcd", "EF")]
            )
        else:
            # case insensitive filesystem
            self.assertSequenceEqual(
                self.glob("a", "bcd", "E*"),
                [self.normpath("a", "bcd", "EF"), self.normpath("a", "bcd", "efg")],
            )
        self.assertSequenceEqual(
            self.glob("a", "bcd", "*g"), [self.normpath("a", "bcd", "efg")]
        )

    def test_glob_directory_names(self):
        self.assertSequenceEqual(self.glob("*", "D"), [self.normpath("a", "D")])
        self.assertSequenceEqual(self.glob("*", "*a"), [])
        self.assertSequenceEqual(
            self.glob("a", "*", "*", "*a"), [self.normpath("a", "bcd", "efg", "ha")]
        )
        self.assertSequenceEqual(
            self.glob("?a?", "*F"),
            map(self.normpath, [os.path.join("aaa", "zzzF"), os.path.join("aab", "F")]),
        )

    def test_glob_directory_with_trailing_slash(self):
        # We are verifying that when there is wildcard pattern which
        # ends with os.sep doesn't blow up.
        paths = glob.glob("*" + os.sep)
        self.assertEqual(len(paths), 4)
        self.assertTrue(all(os.sep in path for path in paths))



from django.core.files import temp as tempfile

from pipeline.compressors import SubProcessCompressor
from pipeline.conf import settings


class CSSTidyCompressor(SubProcessCompressor):
    def compress_css(self, css):
        output_file = tempfile.NamedTemporaryFile(suffix=".pipeline")

        command = (
            settings.CSSTIDY_BINARY,
            "-",
            settings.CSSTIDY_ARGUMENTS,
            output_file.name,
        )
        self.execute_command(command, css)

        filtered_css = output_file.read()
        output_file.close()
        return filtered_css



import gzip
from io import BytesIO

from django import get_version as django_version
from django.contrib.staticfiles.storage import (
    ManifestStaticFilesStorage,
    StaticFilesStorage,
)
from django.contrib.staticfiles.utils import matches_patterns
from django.core.files.base import File

_CACHED_STATIC_FILES_STORAGE_AVAILABLE = django_version() < "3.1"

if _CACHED_STATIC_FILES_STORAGE_AVAILABLE:
    from django.contrib.staticfiles.storage import CachedStaticFilesStorage


class PipelineMixin:
    packing = True

    def post_process(self, paths, dry_run=False, **options):
        if dry_run:
            return

        from pipeline.packager import Packager  # noqa: PLC0415

        packager = Packager(storage=self)
        for package_name in packager.packages["css"]:
            package = packager.package_for("css", package_name)
            output_file = package.output_filename
            if self.packing:
                packager.pack_stylesheets(package)
            paths[output_file] = (self, output_file)
            yield output_file, output_file, True
        for package_name in packager.packages["js"]:
            package = packager.package_for("js", package_name)
            output_file = package.output_filename
            if self.packing:
                packager.pack_javascripts(package)
            paths[output_file] = (self, output_file)
            yield output_file, output_file, True

        super_class = super()
        if hasattr(super_class, "post_process"):
            yield from super_class.post_process(paths.copy(), dry_run, **options)

    def get_available_name(self, name, max_length=None):
        if self.exists(name):
            self.delete(name)
        return name


class GZIPMixin:
    gzip_patterns = ("*.css", "*.js")

    def _compress(self, original_file):
        content = BytesIO()
        gzip_file = gzip.GzipFile(mode="wb", fileobj=content)
        gzip_file.write(original_file.read())
        gzip_file.close()
        content.seek(0)
        return File(content)

    def post_process(self, paths, dry_run=False, **options):
        super_class = super()
        if hasattr(super_class, "post_process"):
            for name, hashed_name, processed in super_class.post_process(
                paths.copy(), dry_run, **options
            ):
                if hashed_name != name:
                    paths[hashed_name] = (self, hashed_name)
                yield name, hashed_name, processed

        if dry_run:
            return

        for path in paths:
            if path:
                if not matches_patterns(path, self.gzip_patterns):
                    continue
                original_file = self.open(path)
                gzipped_path = f"{path}.gz"
                if self.exists(gzipped_path):
                    self.delete(gzipped_path)
                gzipped_file = self._compress(original_file)
                gzipped_path = self.save(gzipped_path, gzipped_file)
                yield gzipped_path, gzipped_path, True


class NonPackagingMixin:
    packing = False


class PipelineStorage(PipelineMixin, StaticFilesStorage):
    pass


class NonPackagingPipelineStorage(NonPackagingMixin, PipelineStorage):
    pass


if _CACHED_STATIC_FILES_STORAGE_AVAILABLE:

    class PipelineCachedStorage(PipelineMixin, CachedStaticFilesStorage):
        # Deprecated since Django 2.2
        # Removed in Django 3.1
        pass

    class NonPackagingPipelineCachedStorage(NonPackagingMixin, PipelineCachedStorage):
        # Deprecated since Django 2.2
        # Removed in Django 3.1
        pass


class PipelineManifestStorage(PipelineMixin, ManifestStaticFilesStorage):
    pass


class NonPackagingPipelineManifestStorage(
    NonPackagingMixin, ManifestStaticFilesStorage
):
    pass



from pipeline.compressors import CompressorBase


class CssHtmlJsMinifyCompressor(CompressorBase):
    """
    CSS, HTML and JS compressor based on the Python library css-html-js-minify
    (https://pypi.org/project/css-html-js-minify/).
    """

    def compress_css(self, css):
        from css_html_js_minify import css_minify  # noqa: PLC0415

        return css_minify(css)

    def compress_js(self, js):
        from css_html_js_minify import js_minify  # noqa: PLC0415

        return js_minify(js)



from django import get_version as django_version
from django.forms import Media
from django.test import TestCase

from pipeline.forms import PipelineFormMedia

from ..utils import pipeline_settings


@pipeline_settings(
    PIPELINE_COLLECTOR_ENABLED=False,
    STYLESHEETS={
        "styles1": {
            "source_filenames": (
                "pipeline/css/first.css",
                "pipeline/css/second.css",
            ),
            "output_filename": "styles1.min.css",
        },
        "styles2": {
            "source_filenames": ("pipeline/css/unicode.css",),
            "output_filename": "styles2.min.css",
        },
        "print": {
            "source_filenames": ("pipeline/css/urls.css",),
            "output_filename": "print.min.css",
        },
    },
    JAVASCRIPT={
        "scripts1": {
            "source_filenames": (
                "pipeline/js/first.js",
                "pipeline/js/second.js",
            ),
            "output_filename": "scripts1.min.js",
        },
        "scripts2": {
            "source_filenames": ("pipeline/js/application.js",),
            "output_filename": "scripts2.min.js",
        },
    },
)
class PipelineFormMediaTests(TestCase):
    """Unit tests for pipeline.forms.PipelineFormMedia."""

    @pipeline_settings(PIPELINE_ENABLED=True)
    def test_css_packages_with_pipeline_enabled(self):
        """Testing PipelineFormMedia.css_packages with PIPELINE_ENABLED=True"""

        class MyMedia(PipelineFormMedia):
            css_packages = {
                "all": ("styles1", "styles2"),
                "print": ("print",),
            }

            css = {"all": ("extra1.css", "extra2.css")}

        media = Media(MyMedia)

        self.assertEqual(
            MyMedia.css,
            {
                "all": [
                    "extra1.css",
                    "extra2.css",
                    "/static/styles1.min.css",
                    "/static/styles2.min.css",
                ],
                "print": ["/static/print.min.css"],
            },
        )
        self.assertEqual(MyMedia.css, media._css)
        expected_regex = [
            r'<link href="{}"( type="text/css")? media="all" '
            'rel="stylesheet"( /)?>'.format(path)
            for path in (
                "/static/extra1.css",
                "/static/extra2.css",
                "/static/styles1.min.css",
                "/static/styles2.min.css",
            )
        ] + [
            r'<link href="/static/print.min.css" (type="text/css" )?'
            'media="print" rel="stylesheet"( /)?>'
        ]
        for rendered_node, expected_node in zip(media.render_css(), expected_regex):
            self.assertRegex(rendered_node, expected_node)

    @pipeline_settings(PIPELINE_ENABLED=False)
    def test_css_packages_with_pipeline_disabled(self):
        """Testing PipelineFormMedia.css_packages with PIPELINE_ENABLED=False"""

        class MyMedia(PipelineFormMedia):
            css_packages = {
                "all": ("styles1", "styles2"),
                "print": ("print",),
            }

            css = {"all": ("extra1.css", "extra2.css")}

        media = Media(MyMedia)

        self.assertEqual(
            MyMedia.css,
            {
                "all": [
                    "extra1.css",
                    "extra2.css",
                    "pipeline/css/first.css",
                    "pipeline/css/second.css",
                    "pipeline/css/unicode.css",
                ],
                "print": ["pipeline/css/urls.css"],
            },
        )
        self.assertEqual(MyMedia.css, media._css)

        expected_regex = [
            '<link href="{}"( type="text/css")? media="all" '
            'rel="stylesheet"( /)?>'.format(path)
            for path in (
                "/static/extra1.css",
                "/static/extra2.css",
                "/static/pipeline/css/first.css",
                "/static/pipeline/css/second.css",
                "/static/pipeline/css/unicode.css",
            )
        ] + [
            '<link href="/static/pipeline/css/urls.css" (type="text/css" )?'
            'media="print" rel="stylesheet"( /)?>'
        ]
        for rendered_node, expected_node in zip(media.render_css(), expected_regex):
            self.assertRegex(rendered_node, expected_node)

    @pipeline_settings(PIPELINE_ENABLED=True)
    def test_js_packages_with_pipeline_enabled(self):
        """Testing PipelineFormMedia.js_packages with PIPELINE_ENABLED=True"""

        class MyMedia(PipelineFormMedia):
            js_packages = ("scripts1", "scripts2")
            js = ("extra1.js", "extra2.js")

        media = Media(MyMedia)

        if django_version() < "3.1":
            script_tag = '<script type="text/javascript" src="%s"></script>'
        else:
            script_tag = '<script src="%s"></script>'

        self.assertEqual(
            MyMedia.js,
            [
                "extra1.js",
                "extra2.js",
                "/static/scripts1.min.js",
                "/static/scripts2.min.js",
            ],
        )
        self.assertEqual(MyMedia.js, media._js)
        self.assertEqual(
            media.render_js(),
            [
                script_tag % path
                for path in (
                    "/static/extra1.js",
                    "/static/extra2.js",
                    "/static/scripts1.min.js",
                    "/static/scripts2.min.js",
                )
            ],
        )

    @pipeline_settings(PIPELINE_ENABLED=False)
    def test_js_packages_with_pipeline_disabled(self):
        """Testing PipelineFormMedia.js_packages with PIPELINE_ENABLED=False"""

        class MyMedia(PipelineFormMedia):
            js_packages = ("scripts1", "scripts2")
            js = ("extra1.js", "extra2.js")

        media = Media(MyMedia)

        if django_version() < "3.1":
            script_tag = '<script type="text/javascript" src="%s"></script>'
        else:
            script_tag = '<script src="%s"></script>'

        self.assertEqual(
            MyMedia.js,
            [
                "extra1.js",
                "extra2.js",
                "pipeline/js/first.js",
                "pipeline/js/second.js",
                "pipeline/js/application.js",
            ],
        )
        self.assertEqual(MyMedia.js, media._js)
        self.assertEqual(
            media.render_js(),
            [
                script_tag % path
                for path in (
                    "/static/extra1.js",
                    "/static/extra2.js",
                    "/static/pipeline/js/first.js",
                    "/static/pipeline/js/second.js",
                    "/static/pipeline/js/application.js",
                )
            ],
        )



from io import StringIO

import django
from django.contrib.staticfiles import finders
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.management import call_command
from django.test import TestCase
from django.test.utils import modify_settings, override_settings

from pipeline.collector import default_collector
from pipeline.storage import PipelineStorage
from tests.tests.test_compiler import DummyCompiler
from tests.utils import pipeline_settings


class PipelineNoPathStorage(PipelineStorage):
    """Storage without an implemented path method"""

    def path(self, *args):
        raise NotImplementedError()

    def delete(self, *args):
        return

    def exists(self, *args):
        return True

    def save(self, *args):
        return

    def open(self, *args):
        return StringIO()

    def listdir(self, *args):
        return []


class DummyCSSCompiler(DummyCompiler):
    """Handles css files"""

    output_extension = "css"

    def match_file(self, path):
        return path.endswith(".css")


class StorageTest(TestCase):
    def tearDown(self):
        staticfiles_storage._setup()

    @pipeline_settings(JS_COMPRESSOR=None, CSS_COMPRESSOR=None)
    def test_post_process_dry_run(self):
        default_collector.collect()
        processed_files = PipelineStorage().post_process({}, True)
        self.assertEqual(list(processed_files), [])

    @pipeline_settings(
        JS_COMPRESSOR=None,
        CSS_COMPRESSOR=None,
        COMPILERS=["tests.tests.test_storage.DummyCSSCompiler"],
    )
    def test_post_process(self):
        default_collector.collect()
        storage = PipelineStorage()
        processed_files = storage.post_process({})
        self.assertTrue(("screen.css", "screen.css", True) in processed_files)
        self.assertTrue(("scripts.js", "scripts.js", True) in processed_files)

    @override_settings(
        STATICFILES_STORAGE="tests.tests.test_storage.PipelineNoPathStorage",
    )
    @pipeline_settings(
        JS_COMPRESSOR=None,
        CSS_COMPRESSOR=None,
        COMPILERS=["tests.tests.test_storage.DummyCSSCompiler"],
    )
    def test_post_process_no_path(self):
        """
        Test post_process with a storage that doesn't implement the path method.
        """
        staticfiles_storage._setup()
        try:
            call_command("collectstatic", verbosity=0, interactive=False)
        except NotImplementedError:
            self.fail("Received an error running collectstatic")

    @modify_settings(STATICFILES_FINDERS={"append": "pipeline.finders.PipelineFinder"})
    def test_nonexistent_file_pipeline_finder(self):
        path = finders.find("nothing.css")
        self.assertIsNone(path)

    @modify_settings(STATICFILES_FINDERS={"append": "pipeline.finders.PipelineFinder"})
    def test_nonexistent_file_pipeline_finder_find_all(self):
        if django.__version__ < "5.2":
            self.skipTest("Only applicable to Django 5.2 and up")

        path = finders.find("nothing.css", find_all=True)
        self.assertIsNotNone(path)
        self.assertEqual([], path)

    @modify_settings(STATICFILES_FINDERS={"append": "pipeline.finders.PipelineFinder"})
    def test_nonexistent_file_pipeline_finder_all(self):
        if django.__version__ < "6.0":
            self.skipTest("Only applicable to versions of Django before 6.0")

        path = finders.find("nothing.css", all=True)
        self.assertIsNotNone(path)
        self.assertEqual([], path)

    @modify_settings(
        STATICFILES_FINDERS={"append": "pipeline.finders.CachedFileFinder"}
    )
    def test_nonexistent_file_cached_finder(self):
        path = finders.find("nothing.css")
        self.assertIsNone(path)

    @modify_settings(STATICFILES_FINDERS={"append": "pipeline.finders.PipelineFinder"})
    def test_nonexistent_double_extension_file_pipeline_finder(self):
        path = finders.find("app.css.map")
        self.assertIsNone(path)

    @modify_settings(
        STATICFILES_FINDERS={"append": "pipeline.finders.CachedFileFinder"}
    )
    def test_nonexistent_double_extension_file_cached_finder(self):
        path = finders.find("app.css.map")
        self.assertIsNone(path)

    @modify_settings(STATICFILES_FINDERS={"append": "pipeline.finders.ManifestFinder"})
    def test_manifest_finder_finds_stylesheet(self):
        path = finders.find("screen.css")
        self.assertIsNotNone(path)

        path = finders.find("screen.scss")
        self.assertIsNone(path)

    @modify_settings(STATICFILES_FINDERS={"append": "pipeline.finders.ManifestFinder"})
    def test_manifest_finder_finds_all_stylesheet(self):
        paths = finders.find("screen.css", all=True)
        self.assertIsNotNone(paths)
        self.assertEqual(1, len(paths))

        paths = finders.find("screen.scss", all=True)
        self.assertIsNotNone(paths)
        self.assertEqual([], paths)



from itertools import chain
from os.path import normpath

from django.contrib.staticfiles.finders import (
    AppDirectoriesFinder as DjangoAppDirectoriesFinder,
)
from django.contrib.staticfiles.finders import BaseFinder, BaseStorageFinder
from django.contrib.staticfiles.finders import (
    FileSystemFinder as DjangoFileSystemFinder,
)
from django.contrib.staticfiles.finders import find
from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils._os import safe_join

from pipeline.conf import settings


class PipelineFinder(BaseStorageFinder):
    storage = staticfiles_storage

    def find(self, path, **kwargs):
        if not settings.PIPELINE_ENABLED:
            return super().find(path, **kwargs)
        else:
            return []

    def list(self, ignore_patterns):
        return []


class ManifestFinder(BaseFinder):
    def find(self, path, **kwargs):
        """
        Looks for files in PIPELINE.STYLESHEETS and PIPELINE.JAVASCRIPT
        """
        matches = []
        for elem in chain(settings.STYLESHEETS.values(), settings.JAVASCRIPT.values()):
            if normpath(elem["output_filename"]) == normpath(path):
                match = safe_join(settings.PIPELINE_ROOT, path)
                if not kwargs.get("find_all", kwargs.get("all", False)):
                    return match
                matches.append(match)
        return matches

    def list(self, *args):
        return []


class CachedFileFinder(BaseFinder):
    def find(self, path, **kwargs):
        """
        Work out the uncached name of the file and look that up instead
        """
        try:
            start, _, extn = path.rsplit(".", 2)
        except ValueError:
            return []
        path = ".".join((start, extn))
        return find(path, **kwargs) or []

    def list(self, *args):
        return []


class PatternFilterMixin:
    ignore_patterns = []

    def get_ignored_patterns(self):
        return list(set(self.ignore_patterns))

    def list(self, ignore_patterns):
        if ignore_patterns:
            ignore_patterns = ignore_patterns + self.get_ignored_patterns()
        return super().list(ignore_patterns)


class AppDirectoriesFinder(PatternFilterMixin, DjangoAppDirectoriesFinder):
    """
    Like AppDirectoriesFinder, but doesn't return any additional ignored
    patterns.

    This allows us to concentrate/compress our components without dragging
    the raw versions in via collectstatic.
    """

    ignore_patterns = [
        "*.js",
        "*.css",
        "*.less",
        "*.scss",
        "*.styl",
    ]


class FileSystemFinder(PatternFilterMixin, DjangoFileSystemFinder):
    """
    Like FileSystemFinder, but doesn't return any additional ignored patterns

    This allows us to concentrate/compress our components without dragging
    the raw versions in too.
    """

    ignore_patterns = [
        "*.js",
        "*.css",
        "*.less",
        "*.scss",
        "*.styl",
        "*.sh",
        "*.html",
        "*.md",
        "*.markdown",
        "*.php",
        "*.txt",
        "README*",
        "LICENSE*",
        "*examples*",
        "*test*",
        "*bin*",
        "*samples*",
        "*docs*",
        "*build*",
        "*demo*",
        "Makefile*",
        "Gemfile*",
        "node_modules",
    ]



from pipeline.compressors import SubProcessCompressor
from pipeline.conf import settings


class ClosureCompressor(SubProcessCompressor):
    def compress_js(self, js):
        command = (settings.CLOSURE_BINARY, settings.CLOSURE_ARGUMENTS)
        return self.execute_command(command, js)



from pipeline.compressors import SubProcessCompressor
from pipeline.conf import settings


class CSSMinCompressor(SubProcessCompressor):
    def compress_css(self, css):
        command = (settings.CSSMIN_BINARY, settings.CSSMIN_ARGUMENTS)
        return self.execute_command(command, css)



import mimetypes

from django.test import TestCase

from pipeline.utils import guess_type


class UtilTest(TestCase):
    def test_guess_type(self):
        self.assertEqual("text/css", guess_type("stylesheet.css"))
        self.assertEqual("text/coffeescript", guess_type("application.coffee"))
        self.assertEqual("text/less", guess_type("stylesheet.less"))

    def test_mimetypes_are_str(self):
        for ext, mtype in mimetypes.types_map.items():
            self.assertIsInstance(ext, str)
            self.assertIsInstance(mtype, str)



import os

from django.contrib.staticfiles import finders
from django.core.files.storage import FileSystemStorage
from django.test import TestCase

from pipeline.collector import default_collector
from pipeline.finders import PipelineFinder


def local_path(path):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", path))


class CollectorTest(TestCase):
    def tearDown(self):
        super().tearDown()

        default_collector.clear()

    def test_collect(self):
        self.assertEqual(
            set(default_collector.collect()), set(self._get_collectable_files())
        )

    def test_collect_with_files(self):
        self.assertEqual(
            set(
                default_collector.collect(
                    files=[
                        "pipeline/js/first.js",
                        "pipeline/js/second.js",
                    ]
                )
            ),
            {
                "pipeline/js/first.js",
                "pipeline/js/second.js",
            },
        )

    def test_delete_file_with_modified(self):
        list(default_collector.collect())

        storage = FileSystemStorage(local_path("assets"))
        new_mtime = os.path.getmtime(storage.path("js/first.js")) - 1000
        os.utime(
            default_collector.storage.path("pipeline/js/first.js"),
            (new_mtime, new_mtime),
        )

        self.assertTrue(
            default_collector.delete_file(
                "js/first.js", "pipeline/js/first.js", storage
            )
        )

    def test_delete_file_with_unmodified(self):
        list(default_collector.collect(files=["pipeline/js/first.js"]))

        self.assertFalse(
            default_collector.delete_file(
                "js/first.js",
                "pipeline/js/first.js",
                FileSystemStorage(local_path("assets")),
            )
        )

    def _get_collectable_files(self):
        for finder in finders.get_finders():
            if not isinstance(finder, PipelineFinder):
                for path, storage in finder.list(["CVS", ".*", "*~"]):
                    if getattr(storage, "prefix", None):
                        yield os.path.join(storage.prefix, path)
                    else:
                        yield path



from pipeline.compressors import SubProcessCompressor
from pipeline.conf import settings


class YuglifyCompressor(SubProcessCompressor):
    def compress_common(self, content, compress_type, arguments):
        command = (settings.YUGLIFY_BINARY, f"--type={compress_type}", arguments)
        return self.execute_command(command, content)

    def compress_js(self, js):
        return self.compress_common(js, "js", settings.YUGLIFY_JS_ARGUMENTS)

    def compress_css(self, css):
        return self.compress_common(css, "css", settings.YUGLIFY_CSS_ARGUMENTS)



from django.core.exceptions import MiddlewareNotUsed
from django.utils.deprecation import MiddlewareMixin
from django.utils.encoding import DjangoUnicodeDecodeError
from django.utils.html import strip_spaces_between_tags as minify_html

from pipeline.conf import settings


class MinifyHTMLMiddleware(MiddlewareMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not settings.PIPELINE_ENABLED:
            raise MiddlewareNotUsed

    def process_response(self, request, response):
        if (
            response.has_header("Content-Type")
            and "text/html" in response["Content-Type"]
        ):
            try:
                response.content = minify_html(response.content.decode("utf-8").strip())
                response["Content-Length"] = str(len(response.content))
            except DjangoUnicodeDecodeError:
                pass
        return response



import sys
from unittest import skipIf, skipUnless

from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.test import TestCase
from django.test.client import RequestFactory

from pipeline.collector import default_collector
from pipeline.compilers import Compiler, CompilerBase, SubProcessCompiler
from pipeline.exceptions import CompilerError
from pipeline.utils import to_class
from tests.utils import _, pipeline_settings


class FailingCompiler(SubProcessCompiler):
    output_extension = "junk"

    def match_file(self, path):
        return path.endswith(".coffee")

    def compile_file(self, infile, outfile, outdated=False, force=False):
        command = (
            (
                "/usr/bin/env",
                "false",
            ),
        )
        return self.execute_command(command)


class InvalidCompiler(SubProcessCompiler):
    output_extension = "junk"

    def match_file(self, path):
        return path.endswith(".coffee")

    def compile_file(self, infile, outfile, outdated=False, force=False):
        command = (
            ("this-exists-nowhere-as-a-command-and-should-fail",),
            infile,
            outfile,
        )
        return self.execute_command(command)


class CompilerWithEmptyFirstArg(SubProcessCompiler):
    output_extension = "junk"

    def match_file(self, path):
        return path.endswith(".coffee")

    def compile_file(self, infile, outfile, outdated=False, force=False):
        command = (
            ("", "/usr/bin/env", "cat"),
            infile,
        )
        return self.execute_command(command, stdout_captured=outfile)


class CopyingCompiler(SubProcessCompiler):
    output_extension = "junk"

    def match_file(self, path):
        return path.endswith(".coffee")

    def compile_file(self, infile, outfile, outdated=False, force=False):
        command = ("cp", infile, outfile)
        return self.execute_command(command)


class LineNumberingCompiler(SubProcessCompiler):
    output_extension = "junk"

    def match_file(self, path):
        return path.endswith(".coffee")

    def compile_file(self, infile, outfile, outdated=False, force=False):
        command = (
            ("/usr/bin/env", "cat"),
            ("-n",),
            infile,
        )
        return self.execute_command(command, stdout_captured=outfile)


class DummyCompiler(CompilerBase):
    output_extension = "js"

    def match_file(self, path):
        return path.endswith(".coffee")

    def compile_file(self, infile, outfile, outdated=False, force=False):
        return


@pipeline_settings(COMPILERS=["tests.tests.test_compiler.DummyCompiler"])
class DummyCompilerTest(TestCase):
    def setUp(self):
        default_collector.collect()
        self.compiler = Compiler()

    def test_output_path(self):
        compiler_class = self.compiler.compilers[0]
        compiler = compiler_class(
            verbose=self.compiler.verbose,
            storage=self.compiler.storage,
        )
        output_path = compiler.output_path("js/helpers.coffee", "js")
        self.assertEqual(output_path, "js/helpers.js")

    def test_compilers_class(self):
        compilers_class = self.compiler.compilers
        self.assertEqual(compilers_class[0], DummyCompiler)

    def test_compile(self):
        paths = self.compiler.compile(
            [
                _("pipeline/js/dummy.coffee"),
                _("pipeline/js/application.js"),
            ]
        )
        self.assertEqual(
            [_("pipeline/js/dummy.js"), _("pipeline/js/application.js")],
            list(paths),
        )

    def tearDown(self):
        default_collector.clear()


@skipIf(sys.platform.startswith("win"), "requires posix platform")
@pipeline_settings(COMPILERS=["tests.tests.test_compiler.LineNumberingCompiler"])
class CompilerStdoutTest(TestCase):
    def setUp(self):
        default_collector.collect()
        self.compiler = Compiler()

    def test_output_path(self):
        compiler_class = self.compiler.compilers[0]
        compiler = compiler_class(
            verbose=self.compiler.verbose,
            storage=self.compiler.storage,
        )
        output_path = compiler.output_path("js/helpers.coffee", "js")
        self.assertEqual(output_path, "js/helpers.js")

    def test_compile(self):
        paths = self.compiler.compile([_("pipeline/js/dummy.coffee")])
        self.assertEqual([_("pipeline/js/dummy.junk")], list(paths))

    def tearDown(self):
        default_collector.clear()


@skipIf(sys.platform.startswith("win"), "requires posix platform")
@pipeline_settings(COMPILERS=["tests.tests.test_compiler.CopyingCompiler"])
class CompilerSelfWriterTest(TestCase):
    def setUp(self):
        default_collector.collect()
        self.compiler = Compiler()

    def test_output_path(self):
        compiler_class = self.compiler.compilers[0]
        compiler = compiler_class(
            verbose=self.compiler.verbose,
            storage=self.compiler.storage,
        )
        output_path = compiler.output_path("js/helpers.coffee", "js")
        self.assertEqual(output_path, "js/helpers.js")

    def test_compile(self):
        paths = self.compiler.compile([_("pipeline/js/dummy.coffee")])
        default_collector.collect()
        self.assertEqual([_("pipeline/js/dummy.junk")], list(paths))

    def tearDown(self):
        default_collector.clear()


@pipeline_settings(COMPILERS=["tests.tests.test_compiler.CompilerWithEmptyFirstArg"])
class CompilerWithEmptyFirstArgTest(TestCase):
    def setUp(self):
        default_collector.collect()
        self.compiler = Compiler()

    def test_compile(self):
        paths = self.compiler.compile([_("pipeline/js/dummy.coffee")])
        default_collector.collect()
        self.assertEqual([_("pipeline/js/dummy.junk")], list(paths))

    def tearDown(self):
        default_collector.clear()


@pipeline_settings(COMPILERS=["tests.tests.test_compiler.InvalidCompiler"])
class InvalidCompilerTest(TestCase):
    def setUp(self):
        default_collector.collect()
        self.compiler = Compiler()

    def test_compile(self):
        with self.assertRaises(CompilerError) as cm:
            self.compiler.compile([_("pipeline/js/dummy.coffee")])
            e = cm.exception
            self.assertEqual(
                e.command,
                [
                    "this-exists-nowhere-as-a-command-and-should-fail",
                    "pipeline/js/dummy.coffee",
                    "pipeline/js/dummy.junk",
                ],
            )
            self.assertEqual(e.error_output, "")

    def tearDown(self):
        default_collector.clear()


@skipIf(sys.platform.startswith("win"), "requires posix platform")
@pipeline_settings(COMPILERS=["tests.tests.test_compiler.FailingCompiler"])
class FailingCompilerTest(TestCase):
    def setUp(self):
        default_collector.collect()
        self.compiler = Compiler()

    def test_compile(self):
        with self.assertRaises(CompilerError) as cm:
            self.compiler.compile([_("pipeline/js/dummy.coffee")])

            e = cm.exception
            self.assertEqual(e.command, ["/usr/bin/env", "false"])
            self.assertEqual(e.error_output, "")

    def tearDown(self):
        default_collector.clear()


@skipUnless(settings.HAS_NODE, "requires node")
class CompilerImplementation(TestCase):
    def setUp(self):
        self.compiler = Compiler()
        default_collector.collect(RequestFactory().get("/"))

    def tearDown(self):
        default_collector.clear()

    def _test_compiler(self, compiler_cls_str, infile, expected):
        compiler_cls = to_class(compiler_cls_str)
        compiler = compiler_cls(verbose=False, storage=staticfiles_storage)
        infile_path = staticfiles_storage.path(infile)
        outfile_path = compiler.output_path(infile_path, compiler.output_extension)
        compiler.compile_file(_(infile_path), _(outfile_path), force=True)
        with open(outfile_path) as f:
            result = f.read()
        with staticfiles_storage.open(expected, "r") as f:
            expected = f.read()
        self.assertEqual(result, expected)

    def test_sass(self):
        self._test_compiler(
            "pipeline.compilers.sass.SASSCompiler",
            "pipeline/compilers/scss/input.scss",
            "pipeline/compilers/scss/expected.css",
        )

    def test_coffeescript(self):
        self._test_compiler(
            "pipeline.compilers.coffee.CoffeeScriptCompiler",
            "pipeline/compilers/coffee/input.coffee",
            "pipeline/compilers/coffee/expected.js",
        )

    def test_less(self):
        self._test_compiler(
            "pipeline.compilers.less.LessCompiler",
            "pipeline/compilers/less/input.less",
            "pipeline/compilers/less/expected.css",
        )

    def test_es6(self):
        self._test_compiler(
            "pipeline.compilers.es6.ES6Compiler",
            "pipeline/compilers/es6/input.es6",
            "pipeline/compilers/es6/expected.js",
        )

    def test_typescript(self):
        self._test_compiler(
            "pipeline.compilers.typescript.TypeScriptCompiler",
            "pipeline/compilers/typescript/input.ts",
            "pipeline/compilers/typescript/expected.js",
        )

    def test_stylus(self):
        self._test_compiler(
            "pipeline.compilers.stylus.StylusCompiler",
            "pipeline/compilers/stylus/input.styl",
            "pipeline/compilers/stylus/expected.css",
        )

    def test_livescript(self):
        self._test_compiler(
            "pipeline.compilers.livescript.LiveScriptCompiler",
            "pipeline/compilers/livescript/input.ls",
            "pipeline/compilers/livescript/expected.js",
        )



import os
import shlex
from collections.abc import MutableMapping

from django.conf import settings as _settings
from django.core.signals import setting_changed
from django.dispatch import receiver

DEFAULTS = {
    "PIPELINE_ENABLED": not _settings.DEBUG,
    "PIPELINE_COLLECTOR_ENABLED": True,
    "PIPELINE_ROOT": _settings.STATIC_ROOT,
    "PIPELINE_URL": _settings.STATIC_URL,
    "SHOW_ERRORS_INLINE": _settings.DEBUG,
    "CSS_COMPRESSOR": "pipeline.compressors.yuglify.YuglifyCompressor",
    "JS_COMPRESSOR": "pipeline.compressors.yuglify.YuglifyCompressor",
    "COMPILERS": [],
    "STYLESHEETS": {},
    "JAVASCRIPT": {},
    "TEMPLATE_NAMESPACE": "window.JST",
    "TEMPLATE_EXT": ".jst",
    "TEMPLATE_FUNC": "template",
    "TEMPLATE_SEPARATOR": "_",
    "DISABLE_WRAPPER": False,
    "JS_WRAPPER": "(function() {\n%s\n}).call(this);",
    "CSSTIDY_BINARY": "/usr/bin/env csstidy",
    "CSSTIDY_ARGUMENTS": "--template=highest",
    "YUGLIFY_BINARY": "/usr/bin/env yuglify",
    "YUGLIFY_CSS_ARGUMENTS": "--terminal",
    "YUGLIFY_JS_ARGUMENTS": "--terminal",
    "YUI_BINARY": "/usr/bin/env yuicompressor",
    "YUI_CSS_ARGUMENTS": "",
    "YUI_JS_ARGUMENTS": "",
    "CLOSURE_BINARY": "/usr/bin/env closure",
    "CLOSURE_ARGUMENTS": "",
    "UGLIFYJS_BINARY": "/usr/bin/env uglifyjs",
    "UGLIFYJS_ARGUMENTS": "",
    "TERSER_BINARY": "/usr/bin/env terser",
    "TERSER_ARGUMENTS": "--compress",
    "CSSMIN_BINARY": "/usr/bin/env cssmin",
    "CSSMIN_ARGUMENTS": "",
    "COFFEE_SCRIPT_BINARY": "/usr/bin/env coffee",
    "COFFEE_SCRIPT_ARGUMENTS": "",
    "BABEL_BINARY": "/usr/bin/env babel",
    "BABEL_ARGUMENTS": "",
    "LIVE_SCRIPT_BINARY": "/usr/bin/env lsc",
    "LIVE_SCRIPT_ARGUMENTS": "",
    "TYPE_SCRIPT_BINARY": "/usr/bin/env tsc",
    "TYPE_SCRIPT_ARGUMENTS": "",
    "SASS_BINARY": "/usr/bin/env sass",
    "SASS_ARGUMENTS": "",
    "STYLUS_BINARY": "/usr/bin/env stylus",
    "STYLUS_ARGUMENTS": "",
    "LESS_BINARY": "/usr/bin/env lessc",
    "LESS_ARGUMENTS": "",
    "MIMETYPES": (
        (("text/coffeescript"), (".coffee")),
        (("text/less"), (".less")),
        (("text/javascript"), (".js")),
        (("text/typescript"), (".ts")),
        (("text/x-sass"), (".sass")),
        (("text/x-scss"), (".scss")),
    ),
    "EMBED_MAX_IMAGE_SIZE": 32700,
    "EMBED_PATH": r"[/]?embed/",
}


class PipelineSettings(MutableMapping):
    """
    Container object for pipeline settings
    """

    def __init__(self, wrapped_settings):
        self.settings = DEFAULTS.copy()
        self.settings.update(wrapped_settings)

    def __getitem__(self, key):
        value = self.settings[key]
        if key.endswith(("_BINARY", "_ARGUMENTS")):
            if isinstance(value, (str,)):
                return tuple(shlex.split(value, posix=(os.name == "posix")))
            return tuple(value)
        return value

    def __setitem__(self, key, value):
        self.settings[key] = value

    def __delitem__(self, key):
        del self.store[key]

    def __iter__(self):
        return iter(self.settings)

    def __len__(self):
        return len(self.settings)

    def __getattr__(self, name):
        return self.__getitem__(name)


settings = PipelineSettings(_settings.PIPELINE)


@receiver(setting_changed)
def reload_settings(**kwargs):
    if kwargs["setting"] == "PIPELINE":
        settings.update(kwargs["value"])



import os
import sys

if sys.platform.startswith("win"):
    os.environ.setdefault("NUMBER_OF_PROCESSORS", "1")


from .test_collector import *  # noqa
from .test_compiler import *  # noqa
from .test_compressor import *  # noqa
from .test_glob import *  # noqa
from .test_middleware import *  # noqa
from .test_packager import *  # noqa
from .test_storage import *  # noqa
from .test_template import *  # noqa
from .test_utils import *  # noqa
from .test_views import *  # noqa



from django.contrib.staticfiles.finders import find, get_finders
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.files.base import ContentFile
from django.utils.encoding import smart_bytes

from pipeline.compilers import Compiler
from pipeline.compressors import Compressor
from pipeline.conf import settings
from pipeline.exceptions import PackageNotFound
from pipeline.glob import glob
from pipeline.signals import css_compressed, js_compressed


class Package:
    def __init__(self, config):
        self.config = config
        self._sources = []

    @property
    def sources(self):
        if not self._sources:
            paths = []
            for pattern in self.config.get("source_filenames", []):
                for path in glob(pattern):
                    if path not in paths and find(path):
                        paths.append(str(path))
            self._sources = paths
        return self._sources

    @property
    def paths(self):
        return [
            path for path in self.sources if not path.endswith(settings.TEMPLATE_EXT)
        ]

    @property
    def templates(self):
        return [path for path in self.sources if path.endswith(settings.TEMPLATE_EXT)]

    @property
    def output_filename(self):
        return self.config.get("output_filename")

    @property
    def extra_context(self):
        return self.config.get("extra_context", {})

    @property
    def template_name(self):
        return self.config.get("template_name")

    @property
    def variant(self):
        return self.config.get("variant")

    @property
    def manifest(self):
        return self.config.get("manifest", True)

    @property
    def compiler_options(self):
        return self.config.get("compiler_options", {})


class Packager:
    def __init__(
        self,
        storage=None,
        verbose=False,
        css_packages=None,
        js_packages=None,
    ):
        if storage is None:
            storage = staticfiles_storage
        self.storage = storage
        self.verbose = verbose
        self.compressor = Compressor(storage=storage, verbose=verbose)
        self.compiler = Compiler(storage=storage, verbose=verbose)
        if css_packages is None:
            css_packages = settings.STYLESHEETS
        if js_packages is None:
            js_packages = settings.JAVASCRIPT
        self.packages = {
            "css": self.create_packages(css_packages),
            "js": self.create_packages(js_packages),
        }

    def package_for(self, kind, package_name):
        try:
            return self.packages[kind][package_name]
        except KeyError:
            raise PackageNotFound(
                "No corresponding package for {} package name : {}".format(
                    kind, package_name
                )
            )

    def individual_url(self, filename):
        return self.storage.url(filename)

    def pack_stylesheets(self, package, **kwargs):
        return self.pack(
            package,
            self.compressor.compress_css,
            css_compressed,
            output_filename=package.output_filename,
            variant=package.variant,
            **kwargs,
        )

    def compile(self, paths, compiler_options={}, force=False):
        paths = self.compiler.compile(
            paths,
            compiler_options=compiler_options,
            force=force,
        )
        for path in paths:
            if not self.storage.exists(path):
                if self.verbose:
                    e = (
                        "Compiled file '%s' cannot be "
                        "found with packager's storage. Locating it."
                    )
                    print(e % path)

                source_storage = self.find_source_storage(path)
                if source_storage is not None:
                    with source_storage.open(path) as source_file:
                        if self.verbose:
                            print(f"Saving: {path}")
                        self.storage.save(path, source_file)
                else:
                    raise OSError(f"File does not exist: {path}")
        return paths

    def pack(self, package, compress, signal, **kwargs):
        output_filename = package.output_filename
        if self.verbose:
            print(f"Saving: {output_filename}")
        paths = self.compile(
            package.paths,
            compiler_options=package.compiler_options,
            force=True,
        )
        content = compress(paths, **kwargs)
        self.save_file(output_filename, content)
        signal.send(sender=self, package=package, **kwargs)
        return output_filename

    def pack_javascripts(self, package, **kwargs):
        return self.pack(
            package,
            self.compressor.compress_js,
            js_compressed,
            output_filename=package.output_filename,
            templates=package.templates,
            **kwargs,
        )

    def pack_templates(self, package):
        return self.compressor.compile_templates(package.templates)

    def save_file(self, path, content):
        return self.storage.save(path, ContentFile(smart_bytes(content)))

    def find_source_storage(self, path):
        for finder in get_finders():
            for short_path, storage in finder.list(""):
                if short_path == path:
                    if self.verbose:
                        print(f"Found storage: {str(self.storage)}")
                    return storage
        return None

    def create_packages(self, config):
        packages = {}
        if not config:
            return packages
        for name in config:
            packages[name] = Package(config[name])
        return packages



"""Support for referencing Pipeline packages in forms and widgets."""

from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils.functional import cached_property

from .collector import default_collector
from .conf import settings
from .packager import Packager


class PipelineFormMediaProperty:
    """A property that converts Pipeline packages to lists of files.

    This is used behind the scenes for any Media classes that subclass
    :py:class:`PipelineFormMedia`. When accessed, it converts any Pipeline
    packages into lists of media files and returns or forwards on lookups to
    that list.
    """

    def __init__(self, get_media_files_func, media_cls, extra_files):
        """Initialize the property.

        Args:
            get_media_files_func (callable):
                The function to call to generate the media files.

            media_cls (type):
                The Media class owning the property.

            extra_files (object):
                Files listed in the original ``css`` or ``js`` attribute on
                the Media class.
        """
        self._get_media_files_func = get_media_files_func
        self._media_cls = media_cls
        self._extra_files = extra_files

    @cached_property
    def _media_files(self):
        """The media files represented by the property."""
        return self._get_media_files_func(self._media_cls, self._extra_files)

    def __get__(self, *args, **kwargs):
        """Return the media files when accessed as an attribute.

        This is called when accessing the attribute directly through the
        Media class (for example, ``Media.css``). It returns the media files
        directly.

        Args:
            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            object:
            The list or dictionary containing the media files definition.
        """
        return self._media_files

    def __getattr__(self, attr_name):
        """Return an attribute on the media files definition.

        This is called when accessing an attribute that doesn't otherwise
        exist in the property's dictionary. The call is forwarded onto the
        media files definition.

        Args:
            attr_name (unicode):
                The attribute name.

        Returns:
            object:
            The attribute value.

        Raises:
            AttributeError:
                An attribute with this name could not be found.
        """
        return getattr(self._media_files, attr_name)

    def __iter__(self):
        """Iterate through the media files definition.

        This is called when attempting to iterate over this property. It
        iterates over the media files definition instead.

        Yields:
            object:
            Each entry in the media files definition.
        """
        return iter(self._media_files)


class PipelineFormMediaMetaClass(type):
    """Metaclass for the PipelineFormMedia class.

    This is responsible for converting CSS/JavaScript packages defined in
    Pipeline into lists of files to include on a page. It handles access to the
    :py:attr:`css` and :py:attr:`js` attributes on the class, generating a
    list of files to return based on the Pipelined packages and individual
    files listed in the :py:attr:`css`/:py:attr:`css_packages` or
    :py:attr:`js`/:py:attr:`js_packages` attributes.
    """

    def __new__(cls, name, bases, attrs):
        """Construct the class.

        Args:
            name (bytes):
                The name of the class.

            bases (tuple):
                The base classes for the class.

            attrs (dict):
                The attributes going into the class.

        Returns:
            type:
            The new class.
        """
        new_class = super().__new__(cls, name, bases, attrs)

        # If we define any packages, we'll need to use our special
        # PipelineFormMediaProperty class. We use this instead of intercepting
        # in __getattribute__ because Django does not access them through
        # normal property access. Instead, grabs the Media class's __dict__
        # and accesses them from there. By using these special properties, we
        # can handle direct access (Media.css) and dictionary-based access
        # (Media.__dict__['css']).
        if "css_packages" in attrs:
            new_class.css = PipelineFormMediaProperty(
                cls._get_css_files, new_class, attrs.get("css") or {}
            )

        if "js_packages" in attrs:
            new_class.js = PipelineFormMediaProperty(
                cls._get_js_files, new_class, attrs.get("js") or []
            )

        return new_class

    def _get_css_files(cls, extra_files):
        """Return all CSS files from the Media class.

        Args:
            extra_files (dict):
                The contents of the Media class's original :py:attr:`css`
                attribute, if one was provided.

        Returns:
            dict:
            The CSS media types and files to return for the :py:attr:`css`
            attribute.
        """
        packager = Packager()
        css_packages = getattr(cls, "css_packages", {})

        return {
            media_target: cls._get_media_files(
                packager=packager,
                media_packages=media_packages,
                media_type="css",
                extra_files=extra_files.get(media_target, []),
            )
            for media_target, media_packages in css_packages.items()
        }

    def _get_js_files(cls, extra_files):
        """Return all JavaScript files from the Media class.

        Args:
            extra_files (list):
                The contents of the Media class's original :py:attr:`js`
                attribute, if one was provided.

        Returns:
            list:
            The JavaScript files to return for the :py:attr:`js` attribute.
        """
        return cls._get_media_files(
            packager=Packager(),
            media_packages=getattr(cls, "js_packages", {}),
            media_type="js",
            extra_files=extra_files,
        )

    def _get_media_files(cls, packager, media_packages, media_type, extra_files):
        """Return source or output media files for a list of packages.

        This will go through the media files belonging to the provided list
        of packages referenced in a Media class and return the output files
        (if Pipeline is enabled) or the source files (if not enabled).

        Args:
            packager (pipeline.packager.Packager):
                The packager responsible for media compilation for this type
                of package.

            media_packages (list of unicode):
                The list of media packages referenced in Media to compile or
                return.

            extra_files (list of unicode):
                The list of extra files to include in the result. This would
                be the list stored in the Media class's original :py:attr:`css`
                or :py:attr:`js` attributes.

        Returns:
            list:
            The list of media files for the given packages.
        """
        source_files = list(extra_files)

        if not settings.PIPELINE_ENABLED and settings.PIPELINE_COLLECTOR_ENABLED:
            default_collector.collect()

        for media_package in media_packages:
            package = packager.package_for(media_type, media_package)

            if settings.PIPELINE_ENABLED:
                source_files.append(staticfiles_storage.url(package.output_filename))
            else:
                source_files += packager.compile(package.paths)

        return source_files


class PipelineFormMedia(metaclass=PipelineFormMediaMetaClass):
    """Base class for form or widget Media classes that use Pipeline packages.

    Forms or widgets that need custom CSS or JavaScript media on a page can
    define a standard :py:class:`Media` class that subclasses this class,
    listing the CSS or JavaScript packages in :py:attr:`css_packages` and
    :py:attr:`js_packages` attributes. These are formatted the same as the
    standard :py:attr:`css` and :py:attr:`js` attributes, but reference
    Pipeline package names instead of individual source files.

    If Pipeline is enabled, these will expand to the output files for the
    packages. Otherwise, these will expand to the list of source files for the
    packages.

    Subclasses can also continue to define :py:attr:`css` and :py:attr:`js`
    attributes, which will be returned along with the other output/source
    files.

    Example:

        from django import forms
        from pipeline.forms import PipelineFormMedia


        class MyForm(forms.Media):
            ...

            class Media(PipelineFormMedia):
                css_packages = {
                    'all': ('my-form-styles-package',
                            'other-form-styles-package'),
                    'print': ('my-form-print-styles-package',),
                }

                js_packages = ('my-form-scripts-package',)
                js = ('some-file.js',)
    """



import glob
import os
import shutil


def local_path(path):
    return os.path.join(os.path.dirname(__file__), path)


DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "TEST_NAME": ":memory:"}
}

DEBUG = False

SITE_ID = 1

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.auth",
    "django.contrib.admin",
    "pipeline",
    "tests.tests",
]


ROOT_URLCONF = "tests.urls"

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
]

MEDIA_URL = "/media/"

MEDIA_ROOT = local_path("media")

STATICFILES_STORAGE = "pipeline.storage.PipelineStorage"
STATIC_ROOT = local_path("static/")
STATIC_URL = "/static/"
STATICFILES_DIRS = (("pipeline", local_path("assets/")),)
STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "pipeline.finders.PipelineFinder",
)

SECRET_KEY = "django-pipeline"

PIPELINE = {
    "PIPELINE_ENABLED": True,
    "JS_COMPRESSOR": None,
    "CSS_COMPRESSOR": None,
    "STYLESHEETS": {
        "screen": {
            "source_filenames": (
                "pipeline/css/first.css",
                "pipeline/css/second.css",
                "pipeline/css/urls.css",
            ),
            "output_filename": "screen.css",
        },
        "screen_media": {
            "source_filenames": (
                "pipeline/css/first.css",
                "pipeline/css/second.css",
                "pipeline/css/urls.css",
            ),
            "output_filename": "screen_media.css",
            "extra_context": {
                "media": "screen and (min-width:500px)",
            },
        },
        "screen_title": {
            "source_filenames": (
                "pipeline/css/first.css",
                "pipeline/css/second.css",
                "pipeline/css/urls.css",
            ),
            "output_filename": "screen_title.css",
            "extra_context": {
                "title": "Default Style",
            },
        },
    },
    "JAVASCRIPT": {
        "scripts": {
            "source_filenames": (
                "pipeline/js/first.js",
                "pipeline/js/second.js",
                "pipeline/js/application.js",
                "pipeline/templates/**/*.jst",
            ),
            "output_filename": "scripts.js",
        },
        "scripts_async": {
            "source_filenames": (
                "pipeline/js/first.js",
                "pipeline/js/second.js",
                "pipeline/js/application.js",
                "pipeline/templates/**/*.jst",
            ),
            "output_filename": "scripts_async.js",
            "extra_context": {
                "async": True,
            },
        },
        "scripts_defer": {
            "source_filenames": (
                "pipeline/js/first.js",
                "pipeline/js/second.js",
                "pipeline/js/application.js",
                "pipeline/templates/**/*.jst",
            ),
            "output_filename": "scripts_defer.js",
            "extra_context": {
                "defer": True,
            },
        },
        "scripts_async_defer": {
            "source_filenames": (
                "pipeline/js/first.js",
                "pipeline/js/second.js",
                "pipeline/js/application.js",
                "pipeline/templates/**/*.jst",
            ),
            "output_filename": "scripts_async_defer.js",
            "extra_context": {
                "async": True,
                "defer": True,
            },
        },
    },
}

NODE_MODULES_PATH = local_path("../node_modules")
NODE_BIN_PATH = os.path.join(NODE_MODULES_PATH, ".bin")
NODE_EXE_PATH = shutil.which("node")
JAVA_EXE_PATH = shutil.which("java")
CSSTIDY_EXE_PATH = shutil.which("csstidy")
HAS_NODE = bool(NODE_EXE_PATH)
HAS_JAVA = bool(JAVA_EXE_PATH)
HAS_CSSTIDY = bool(CSSTIDY_EXE_PATH)

if HAS_NODE:

    def node_exe_path(command):
        exe_ext = ".cmd" if os.name == "nt" else ""
        return os.path.join(NODE_BIN_PATH, "{}{}".format(command, exe_ext))

    PIPELINE.update(
        {
            "SASS_BINARY": node_exe_path("sass"),
            "COFFEE_SCRIPT_BINARY": node_exe_path("coffee"),
            "COFFEE_SCRIPT_ARGUMENTS": ["--no-header"],
            "LESS_BINARY": node_exe_path("lessc"),
            "BABEL_BINARY": node_exe_path("babel"),
            "BABEL_ARGUMENTS": ["--presets", "es2015"],
            "STYLUS_BINARY": node_exe_path("stylus"),
            "LIVE_SCRIPT_BINARY": node_exe_path("lsc"),
            "LIVE_SCRIPT_ARGUMENTS": ["--no-header"],
            "YUGLIFY_BINARY": node_exe_path("yuglify"),
            "UGLIFYJS_BINARY": node_exe_path("uglifyjs"),
            "TERSER_BINARY": node_exe_path("terser"),
            "CSSMIN_BINARY": node_exe_path("cssmin"),
            "TYPE_SCRIPT_BINARY": node_exe_path("tsc"),
        }
    )

if HAS_NODE and HAS_JAVA:
    PIPELINE.update(
        {
            "CLOSURE_BINARY": [
                JAVA_EXE_PATH,
                "-jar",
                os.path.join(
                    NODE_MODULES_PATH,
                    "google-closure-compiler-java",
                    "compiler.jar",
                ),
            ],
            "YUI_BINARY": [
                JAVA_EXE_PATH,
                "-jar",
                glob.glob(
                    os.path.join(NODE_MODULES_PATH, "yuicompressor", "build", "*.jar")
                )[0],
            ],
        }
    )

if HAS_CSSTIDY:
    PIPELINE.update({"CSSTIDY_BINARY": CSSTIDY_EXE_PATH})

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "DIRS": [local_path("templates")],
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    },
    {
        "BACKEND": "django.template.backends.jinja2.Jinja2",
        "APP_DIRS": True,
        "DIRS": [local_path("templates")],
        "OPTIONS": {"extensions": ["pipeline.jinja2.PipelineExtension"]},
    },
]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "pipeline.templatetags.pipeline": {
            "handlers": ["console"],
            "level": "ERROR",
        },
    },
}



from pipeline.compressors import SubProcessCompressor
from pipeline.conf import settings


class UglifyJSCompressor(SubProcessCompressor):
    def compress_js(self, js):
        command = (settings.UGLIFYJS_BINARY, settings.UGLIFYJS_ARGUMENTS)
        if self.verbose:
            command += " --verbose"
        return self.execute_command(command, js)



from os.path import dirname

from pipeline.compilers import SubProcessCompiler
from pipeline.conf import settings


class LessCompiler(SubProcessCompiler):
    output_extension = "css"

    def match_file(self, filename):
        return filename.endswith(".less")

    def compile_file(self, infile, outfile, outdated=False, force=False):
        # Pipe to file rather than provide outfile arg due to a bug in lessc
        command = (
            settings.LESS_BINARY,
            settings.LESS_ARGUMENTS,
            infile,
        )
        return self.execute_command(
            command, cwd=dirname(infile), stdout_captured=outfile
        )



PackageNotFoundError = None
DistributionNotFound = None
try:
    from importlib.metadata import PackageNotFoundError
    from importlib.metadata import version as get_version
except ImportError:
    get_version = None
    PackageNotFoundError = None
if get_version is None:
    try:
        from pkg_resources import DistributionNotFound, get_distribution

        def get_version(x):
            return get_distribution(x).version

    except ImportError:
        get_version = None
        DistributionNotFound = None
        get_distribution = None

__version__ = None
if get_version is not None:
    try:
        __version__ = get_version("django-pipeline")
    except PackageNotFoundError:
        pass
    except DistributionNotFound:
        pass



import base64
import os
import sys

try:
    from unittest.mock import patch
except ImportError:
    from unittest.mock import patch  # noqa

from unittest import skipIf, skipUnless

from django.conf import settings
from django.test import TestCase
from django.test.client import RequestFactory

from pipeline.collector import default_collector
from pipeline.compressors import (
    CSS_REWRITE_PATH_RE,
    JS_REWRITE_PATH_RE,
    TEMPLATE_FUNC,
    Compressor,
    SubProcessCompressor,
)
from pipeline.compressors.yuglify import YuglifyCompressor
from tests.utils import _, pipeline_settings


@pipeline_settings(
    CSS_COMPRESSOR="pipeline.compressors.yuglify.YuglifyCompressor",
    JS_COMPRESSOR="pipeline.compressors.yuglify.YuglifyCompressor",
)
class CompressorTest(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.compressor = Compressor()
        default_collector.collect()

    def test_js_compressor_class(self):
        self.assertEqual(self.compressor.js_compressor, YuglifyCompressor)

    def test_css_compressor_class(self):
        self.assertEqual(self.compressor.css_compressor, YuglifyCompressor)

    def test_concatenate_and_rewrite(self):
        css = self.compressor.concatenate_and_rewrite(
            [_("pipeline/css/first.css"), _("pipeline/css/second.css")],
            "css/screen.css",
        )
        expected = """.concat {\n  display: none;\n}\n\n.concatenate {\n  display: block;\n}\n"""  # noqa
        self.assertEqual(expected, css)

    def test_concatenate(self):
        js = self.compressor.concatenate(
            [_("pipeline/js/first.js"), _("pipeline/js/second.js")]
        )
        expected = """(function() {\n  window.concat = function() {\n    console.log(arguments);\n  }\n}()) // No semicolon\n\n;(function() {\n  window.cat = function() {\n    console.log("hello world");\n  }\n}());\n"""  # noqa
        self.assertEqual(expected, js)

    @patch.object(base64, "b64encode")
    def test_encoded_content(self, mock):
        self.compressor.asset_contents.clear()
        self.compressor.encoded_content(_("pipeline/images/arrow.png"))
        self.assertTrue(mock.called)
        mock.reset_mock()
        self.compressor.encoded_content(_("pipeline/images/arrow.png"))
        self.assertFalse(mock.called)

    def test_encoded_content_output(self):
        self.compressor.asset_contents.clear()
        encoded = self.compressor.encoded_content(_("pipeline/images/arrow.png"))
        expected = (
            "iVBORw0KGgoAAAANSUhEUgAAAAkAAAAGCAYAAAARx7TFAAAAMk"
            "lEQVR42oXKwQkAMAxC0Q7rEk5voSEepCHC9/SOpLV3JPULgArV"
            "RtDIMEEiQ4NECRNdciCfK3K3wvEAAAAASUVORK5CYII="
        )
        self.assertEqual(encoded, expected)

    def test_relative_path(self):
        relative_path = self.compressor.relative_path(
            "images/sprite.png",
            "css/screen.css",
        )
        self.assertEqual(relative_path, "../images/sprite.png")

    def test_base_path(self):
        base_path = self.compressor.base_path(
            [_("js/templates/form.jst"), _("js/templates/field.jst")]
        )
        self.assertEqual(base_path, _("js/templates"))

    def test_absolute_path(self):
        absolute_path = self.compressor.absolute_path(
            "../../images/sprite.png", "css/plugins/"
        )
        self.assertEqual(absolute_path, "images/sprite.png")
        absolute_path = self.compressor.absolute_path(
            "/images/sprite.png", "css/plugins/"
        )
        self.assertEqual(absolute_path, "/images/sprite.png")

    def test_template_name(self):
        name = self.compressor.template_name("templates/photo/detail.jst", "templates/")
        self.assertEqual(name, "photo_detail")
        name = self.compressor.template_name("templates/photo_edit.jst", "")
        self.assertEqual(name, "photo_edit")
        name = self.compressor.template_name(
            r"templates\photo\detail.jst",  # noqa
            "templates\\",
        )
        self.assertEqual(name, "photo_detail")

    @pipeline_settings(TEMPLATE_SEPARATOR="/")
    def test_template_name_separator(self):
        name = self.compressor.template_name("templates/photo/detail.jst", "templates/")
        self.assertEqual(name, "photo/detail")
        name = self.compressor.template_name("templates/photo_edit.jst", "")
        self.assertEqual(name, "photo_edit")
        name = self.compressor.template_name(
            r"templates\photo\detail.jst",  # noqa
            "templates\\",
        )
        self.assertEqual(name, "photo/detail")

    def test_compile_templates(self):
        templates = self.compressor.compile_templates(
            [_("pipeline/templates/photo/list.jst")]
        )
        self.assertEqual(
            templates,
            """window.JST = window.JST || {};\n%s\nwindow.JST[\'list\'] = template(\'<div class="photo">\\n <img src="<%%= src %%>" />\\n <div class="caption">\\n  <%%= caption %%>\\n </div>\\n</div>\');\n"""  # noqa
            % TEMPLATE_FUNC,
        )
        templates = self.compressor.compile_templates(
            [
                _("pipeline/templates/video/detail.jst"),
                _("pipeline/templates/photo/detail.jst"),
            ]
        )
        self.assertEqual(
            templates,
            """window.JST = window.JST || {};\n%s\nwindow.JST[\'video_detail\'] = template(\'<div class="video">\\n <video src="<%%= src %%>" />\\n <div class="caption">\\n  <%%= description %%>\\n </div>\\n</div>\');\nwindow.JST[\'photo_detail\'] = template(\'<div class="photo">\\n <img src="<%%= src %%>" />\\n <div class="caption">\\n  <%%= caption %%> by <%%= author %%>\\n </div>\\n</div>\');\n"""  # noqa
            % TEMPLATE_FUNC,
        )

    def test_embeddable(self):
        self.assertFalse(
            self.compressor.embeddable(_("pipeline/images/sprite.png"), None)
        )
        self.assertFalse(
            self.compressor.embeddable(_("pipeline/images/arrow.png"), "datauri")
        )
        self.assertTrue(
            self.compressor.embeddable(_("pipeline/images/embed/arrow.png"), "datauri")
        )
        self.assertFalse(
            self.compressor.embeddable(_("pipeline/images/arrow.dat"), "datauri")
        )

    def test_construct_asset_path(self):
        asset_path = self.compressor.construct_asset_path(
            "../../images/sprite.png", "css/plugins/gallery.css", "css/gallery.css"
        )
        self.assertEqual(asset_path, "../images/sprite.png")
        asset_path = self.compressor.construct_asset_path(
            "/images/sprite.png", "css/plugins/gallery.css", "css/gallery.css"
        )
        self.assertEqual(asset_path, "/images/sprite.png")

    def test_concatenate_with_url_rewrite(self) -> None:
        output = self.compressor.concatenate(
            [
                _("pipeline/css/urls.css"),
            ],
            file_sep="",
            output_filename="css/screen.css",
            rewrite_path_re=CSS_REWRITE_PATH_RE,
        )

        self.assertEqual(
            """.embedded-url-svg {
  background-image: url("data:image/svg+xml;charset=utf8,%3Csvg viewBox='0 0 32 32' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath      stroke='rgba(255, 255, 255, 0.5)' stroke-width='2' stroke-linecap='round' stroke-miterlimit='10' d='M4 8h24M4 16h24M4 24h24'/%3E%     3C/svg%3E");
}
@font-face {
  font-family: 'Pipeline';
  src: url('../pipeline/fonts/pipeline.eot');
  src: url('../pipeline/fonts/pipeline.eot?#iefix') format('embedded-opentype');
  src: local('☺'), url('../pipeline/fonts/pipeline.woff') format('woff'), url('../pipeline/fonts/pipeline.ttf') format('truetype'), url('../pipeline/fonts/pipeline.svg#IyfZbseF') format('svg');
  font-weight: normal;
  font-style: normal;
}
.relative-url {
  background-image: url(../pipeline/images/sprite-buttons.png);
}
.relative-url-querystring {
  background-image: url(../pipeline/images/sprite-buttons.png?v=1.0#foo=bar);
}
.absolute-url {
  background-image: url(/images/sprite-buttons.png);
}
.absolute-full-url {
  background-image: url(http://localhost/images/sprite-buttons.png);
}
.no-protocol-url {
  background-image: url(//images/sprite-buttons.png);
}
.anchor-tag-url {
  background-image: url(#image-gradient);
}
@font-face{src:url(../pipeline/fonts/pipeline.eot);src:url(../pipeline/fonts/pipeline.eot?#iefix) format('embedded-opentype'),url(../pipeline/fonts/pipeline.woff) format('woff'),url(../pipeline/fonts/pipeline.ttf) format('truetype');}
""",  # noqa
            output,
        )

    def test_concatenate_with_url_rewrite_data_uri(self):
        output = self.compressor.concatenate(
            [
                _("pipeline/css/nested/nested.css"),
            ],
            file_sep="",
            output_filename="pipeline/screen.css",
            rewrite_path_re=CSS_REWRITE_PATH_RE,
        )

        self.assertEqual(
            """.data-url {
  background-image: url(data:image/svg+xml;charset=US-ASCII,%3C%3Fxml%20version%3D%221.0%22%20encoding%3D%22iso-8859-1%22%3F%3E%3C!DOCTYPE%20svg%20PUBLIC%20%22-%2F%2FW3C%2F%2FDTD%20SVG%201.1%2F%2FEN%22%20%22http%3A%2F%2Fwww.w3.org%2FGraphics%2FSVG%2F1.1%2FDTD%2Fsvg11.dtd%22%3E%3Csvg%20version%3D%221.1%22%20id%3D%22Layer_1%22%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20xmlns%3Axlink%3D%22http%3A%2F%2Fwww.w3.org%2F1999%2Fxlink%22%20x%3D%220px%22%20y%3D%220px%22%20%20width%3D%2212px%22%20height%3D%2214px%22%20viewBox%3D%220%200%2012%2014%22%20style%3D%22enable-background%3Anew%200%200%2012%2014%3B%22%20xml%3Aspace%3D%22preserve%22%3E%3Cpath%20d%3D%22M11%2C6V5c0-2.762-2.239-5-5-5S1%2C2.238%2C1%2C5v1H0v8h12V6H11z%20M6.5%2C9.847V12h-1V9.847C5.207%2C9.673%2C5%2C9.366%2C5%2C9%20c0-0.553%2C0.448-1%2C1-1s1%2C0.447%2C1%2C1C7%2C9.366%2C6.793%2C9.673%2C6.5%2C9.847z%20M9%2C6H3V5c0-1.657%2C1.343-3%2C3-3s3%2C1.343%2C3%2C3V6z%22%2F%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3C%2Fsvg%3E);
}
.data-url-quoted {
  background-image: url('data:image/svg+xml;charset=US-ASCII,%3C%3Fxml%20version%3D%221.0%22%20encoding%3D%22iso-8859-1%22%3F%3E%3C!DOCTYPE%20svg%20PUBLIC%20%22-%2F%2FW3C%2F%2FDTD%20SVG%201.1%2F%2FEN%22%20%22http%3A%2F%2Fwww.w3.org%2FGraphics%2FSVG%2F1.1%2FDTD%2Fsvg11.dtd%22%3E%3Csvg%20version%3D%221.1%22%20id%3D%22Layer_1%22%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20xmlns%3Axlink%3D%22http%3A%2F%2Fwww.w3.org%2F1999%2Fxlink%22%20x%3D%220px%22%20y%3D%220px%22%20%20width%3D%2212px%22%20height%3D%2214px%22%20viewBox%3D%220%200%2012%2014%22%20style%3D%22enable-background%3Anew%200%200%2012%2014%3B%22%20xml%3Aspace%3D%22preserve%22%3E%3Cpath%20d%3D%22M11%2C6V5c0-2.762-2.239-5-5-5S1%2C2.238%2C1%2C5v1H0v8h12V6H11z%20M6.5%2C9.847V12h-1V9.847C5.207%2C9.673%2C5%2C9.366%2C5%2C9%20c0-0.553%2C0.448-1%2C1-1s1%2C0.447%2C1%2C1C7%2C9.366%2C6.793%2C9.673%2C6.5%2C9.847z%20M9%2C6H3V5c0-1.657%2C1.343-3%2C3-3s3%2C1.343%2C3%2C3V6z%22%2F%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3C%2Fsvg%3E');
}
""",  # noqa
            output,
        )

    def test_concatenate_css_with_sourcemap(self) -> None:
        output = self.compressor.concatenate(
            [
                _("pipeline/css/sourcemap.css"),
            ],
            file_sep="",
            output_filename="css/sourcemap-bundle.css",
            rewrite_path_re=CSS_REWRITE_PATH_RE,
        )

        self.assertEqual(
            output,
            "div {\n"
            "  display: inline;\n"
            "}\n"
            "\n"
            "span {\n"
            "  display: block;\n"
            "}\n"
            "\n"
            "\n"
            "//#  sourceMappingURL=../pipeline/css/sourcemap1.css.map\n"
            "\n"
            "//@ sourceMappingURL=../pipeline/css/sourcemap2.css.map  \n"
            "\n"
            "/*#  sourceMappingURL=../pipeline/css/sourcemap3.css.map */\n"
            "\n"
            "/*@ sourceMappingURL=../pipeline/css/sourcemap4.css.map  */\n"
            "\n"
            "//#  sourceURL=../pipeline/css/sourcemap5.css.map\n"
            "\n"
            "//@ sourceURL=../pipeline/css/sourcemap6.css.map  \n"
            "\n"
            "/*#  sourceURL=../pipeline/css/sourcemap7.css.map */\n"
            "\n"
            "/*@ sourceURL=../pipeline/css/sourcemap8.css.map  */\n",
        )

    def test_concatenate_js_with_sourcemap(self) -> None:
        output = self.compressor.concatenate(
            [
                _("pipeline/js/sourcemap.js"),
            ],
            file_sep=";",
            output_filename="js/sourcemap-bundle.js",
            rewrite_path_re=JS_REWRITE_PATH_RE,
        )

        self.assertEqual(
            output,
            "const abc = 123;\n"
            "\n"
            "\n"
            "//#  sourceMappingURL=../pipeline/js/sourcemap1.js.map\n"
            "\n"
            "//@ sourceMappingURL=../pipeline/js/sourcemap2.js.map  \n"
            "\n"
            "/*#  sourceMappingURL=../pipeline/js/sourcemap3.js.map */\n"
            "\n"
            "/*@ sourceMappingURL=../pipeline/js/sourcemap4.js.map  */\n"
            "\n"
            "//#  sourceURL=../pipeline/js/sourcemap5.js.map\n"
            "\n"
            "//@ sourceURL=../pipeline/js/sourcemap6.js.map  \n"
            "\n"
            "/*#  sourceURL=../pipeline/js/sourcemap7.js.map */\n"
            "\n"
            "/*@ sourceURL=../pipeline/js/sourcemap8.js.map  */\n",
        )

    def test_concatenate_without_rewrite_path_re(self) -> None:
        message = (
            "Compressor.concatenate() was called without passing "
            "rewrite_path_re_= or output_filename=. If you are "
            "specializing Compressor, please update your call "
            "to remain compatible with future changes."
        )

        with self.assertWarnsMessage(DeprecationWarning, message):
            output = self.compressor.concatenate(
                [
                    _("pipeline/js/sourcemap.js"),
                ],
                file_sep=";",
                output_filename="js/sourcemap-bundle.js",
            )

        self.assertEqual(
            output,
            "const abc = 123;\n"
            "\n"
            "\n"
            "//#  sourceMappingURL=sourcemap1.js.map\n"
            "\n"
            "//@ sourceMappingURL=sourcemap2.js.map  \n"
            "\n"
            "/*#  sourceMappingURL=sourcemap3.js.map */\n"
            "\n"
            "/*@ sourceMappingURL=sourcemap4.js.map  */\n"
            "\n"
            "//#  sourceURL=sourcemap5.js.map\n"
            "\n"
            "//@ sourceURL=sourcemap6.js.map  \n"
            "\n"
            "/*#  sourceURL=sourcemap7.js.map */\n"
            "\n"
            "/*@ sourceURL=sourcemap8.js.map  */\n",
        )

    def test_concatenate_without_output_filename(self) -> None:
        message = (
            "Compressor.concatenate() was called without passing "
            "rewrite_path_re_= or output_filename=. If you are "
            "specializing Compressor, please update your call "
            "to remain compatible with future changes."
        )

        with self.assertWarnsMessage(DeprecationWarning, message):
            output = self.compressor.concatenate(
                [
                    _("pipeline/js/sourcemap.js"),
                ],
                file_sep=";",
                rewrite_path_re=JS_REWRITE_PATH_RE,
            )

        self.assertEqual(
            output,
            "const abc = 123;\n"
            "\n"
            "\n"
            "//#  sourceMappingURL=sourcemap1.js.map\n"
            "\n"
            "//@ sourceMappingURL=sourcemap2.js.map  \n"
            "\n"
            "/*#  sourceMappingURL=sourcemap3.js.map */\n"
            "\n"
            "/*@ sourceMappingURL=sourcemap4.js.map  */\n"
            "\n"
            "//#  sourceURL=sourcemap5.js.map\n"
            "\n"
            "//@ sourceURL=sourcemap6.js.map  \n"
            "\n"
            "/*#  sourceURL=sourcemap7.js.map */\n"
            "\n"
            "/*@ sourceURL=sourcemap8.js.map  */\n",
        )

    def test_concatenate_without_file_sep(self) -> None:
        message = (
            "Compressor.concatenate() was called without passing "
            "file_sep=. If you are specializing Compressor, please "
            "update your call to remain compatible with future changes. "
            "Defaulting to JavaScript behavior for "
            "backwards-compatibility."
        )

        with self.assertWarnsMessage(DeprecationWarning, message):
            output = self.compressor.concatenate(
                [
                    _("pipeline/js/first.js"),
                    _("pipeline/js/second.js"),
                ],
                output_filename="js/sourcemap-bundle.js",
                rewrite_path_re=JS_REWRITE_PATH_RE,
            )

        self.assertEqual(
            output,
            "(function() {\n"
            "  window.concat = function() {\n"
            "    console.log(arguments);\n"
            "  }\n"
            "}()) // No semicolon\n"
            "\n"
            ";(function() {\n"
            "  window.cat = function() {\n"
            '    console.log("hello world");\n'
            "  }\n"
            "}());\n",
        )

    def test_legacy_concatenate_and_rewrite(self) -> None:
        message = (
            "Compressor.concatenate_and_rewrite() is deprecated. Please "
            "call concatenate() instead."
        )

        with self.assertWarnsMessage(DeprecationWarning, message):
            output = self.compressor.concatenate_and_rewrite(
                [
                    _("pipeline/css/urls.css"),
                ],
                "css/screen.css",
            )

        self.assertEqual(
            """.embedded-url-svg {
  background-image: url("data:image/svg+xml;charset=utf8,%3Csvg viewBox='0 0 32 32' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath      stroke='rgba(255, 255, 255, 0.5)' stroke-width='2' stroke-linecap='round' stroke-miterlimit='10' d='M4 8h24M4 16h24M4 24h24'/%3E%     3C/svg%3E");
}
@font-face {
  font-family: 'Pipeline';
  src: url('../pipeline/fonts/pipeline.eot');
  src: url('../pipeline/fonts/pipeline.eot?#iefix') format('embedded-opentype');
  src: local('☺'), url('../pipeline/fonts/pipeline.woff') format('woff'), url('../pipeline/fonts/pipeline.ttf') format('truetype'), url('../pipeline/fonts/pipeline.svg#IyfZbseF') format('svg');
  font-weight: normal;
  font-style: normal;
}
.relative-url {
  background-image: url(../pipeline/images/sprite-buttons.png);
}
.relative-url-querystring {
  background-image: url(../pipeline/images/sprite-buttons.png?v=1.0#foo=bar);
}
.absolute-url {
  background-image: url(/images/sprite-buttons.png);
}
.absolute-full-url {
  background-image: url(http://localhost/images/sprite-buttons.png);
}
.no-protocol-url {
  background-image: url(//images/sprite-buttons.png);
}
.anchor-tag-url {
  background-image: url(#image-gradient);
}
@font-face{src:url(../pipeline/fonts/pipeline.eot);src:url(../pipeline/fonts/pipeline.eot?#iefix) format('embedded-opentype'),url(../pipeline/fonts/pipeline.woff) format('woff'),url(../pipeline/fonts/pipeline.ttf) format('truetype');}
""",  # noqa
            output,
        )

    def test_legacy_concatenate_and_rewrite_with_data_uri(self) -> None:
        message = (
            "Compressor.concatenate_and_rewrite() is deprecated. Please "
            "call concatenate() instead."
        )

        with self.assertWarnsMessage(DeprecationWarning, message):
            output = self.compressor.concatenate_and_rewrite(
                [
                    _("pipeline/css/nested/nested.css"),
                ],
                "pipeline/screen.css",
            )

        self.assertEqual(
            """.data-url {
  background-image: url(data:image/svg+xml;charset=US-ASCII,%3C%3Fxml%20version%3D%221.0%22%20encoding%3D%22iso-8859-1%22%3F%3E%3C!DOCTYPE%20svg%20PUBLIC%20%22-%2F%2FW3C%2F%2FDTD%20SVG%201.1%2F%2FEN%22%20%22http%3A%2F%2Fwww.w3.org%2FGraphics%2FSVG%2F1.1%2FDTD%2Fsvg11.dtd%22%3E%3Csvg%20version%3D%221.1%22%20id%3D%22Layer_1%22%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20xmlns%3Axlink%3D%22http%3A%2F%2Fwww.w3.org%2F1999%2Fxlink%22%20x%3D%220px%22%20y%3D%220px%22%20%20width%3D%2212px%22%20height%3D%2214px%22%20viewBox%3D%220%200%2012%2014%22%20style%3D%22enable-background%3Anew%200%200%2012%2014%3B%22%20xml%3Aspace%3D%22preserve%22%3E%3Cpath%20d%3D%22M11%2C6V5c0-2.762-2.239-5-5-5S1%2C2.238%2C1%2C5v1H0v8h12V6H11z%20M6.5%2C9.847V12h-1V9.847C5.207%2C9.673%2C5%2C9.366%2C5%2C9%20c0-0.553%2C0.448-1%2C1-1s1%2C0.447%2C1%2C1C7%2C9.366%2C6.793%2C9.673%2C6.5%2C9.847z%20M9%2C6H3V5c0-1.657%2C1.343-3%2C3-3s3%2C1.343%2C3%2C3V6z%22%2F%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3C%2Fsvg%3E);
}
.data-url-quoted {
  background-image: url('data:image/svg+xml;charset=US-ASCII,%3C%3Fxml%20version%3D%221.0%22%20encoding%3D%22iso-8859-1%22%3F%3E%3C!DOCTYPE%20svg%20PUBLIC%20%22-%2F%2FW3C%2F%2FDTD%20SVG%201.1%2F%2FEN%22%20%22http%3A%2F%2Fwww.w3.org%2FGraphics%2FSVG%2F1.1%2FDTD%2Fsvg11.dtd%22%3E%3Csvg%20version%3D%221.1%22%20id%3D%22Layer_1%22%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20xmlns%3Axlink%3D%22http%3A%2F%2Fwww.w3.org%2F1999%2Fxlink%22%20x%3D%220px%22%20y%3D%220px%22%20%20width%3D%2212px%22%20height%3D%2214px%22%20viewBox%3D%220%200%2012%2014%22%20style%3D%22enable-background%3Anew%200%200%2012%2014%3B%22%20xml%3Aspace%3D%22preserve%22%3E%3Cpath%20d%3D%22M11%2C6V5c0-2.762-2.239-5-5-5S1%2C2.238%2C1%2C5v1H0v8h12V6H11z%20M6.5%2C9.847V12h-1V9.847C5.207%2C9.673%2C5%2C9.366%2C5%2C9%20c0-0.553%2C0.448-1%2C1-1s1%2C0.447%2C1%2C1C7%2C9.366%2C6.793%2C9.673%2C6.5%2C9.847z%20M9%2C6H3V5c0-1.657%2C1.343-3%2C3-3s3%2C1.343%2C3%2C3V6z%22%2F%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3Cg%3E%3C%2Fg%3E%3C%2Fsvg%3E');
}
""",  # noqa
            output,
        )

    @skipIf(sys.platform.startswith("win"), "requires posix platform")
    def test_compressor_subprocess_unicode(self):
        path = os.path.dirname(os.path.dirname(__file__))
        content = open(path + "/assets/css/unicode.css", encoding="utf-8").read()
        output = SubProcessCompressor(False).execute_command(("cat",), content)
        self.assertEqual(
            """.some_class {
  // Some unicode
  content: "áéíóú";
}
""",
            output,
        )

    def tearDown(self):
        default_collector.clear()


class CompressorImplementationTest(TestCase):
    maxDiff = None

    def setUp(self):
        self.compressor = Compressor()
        default_collector.collect(RequestFactory().get("/"))

    def tearDown(self):
        default_collector.clear()

    def _test_compressor(self, compressor_cls, compress_type, expected_file):
        override_settings = {
            (f"{compress_type.upper()}_COMPRESSOR"): compressor_cls,
        }
        with pipeline_settings(**override_settings):
            if compress_type == "js":
                result = self.compressor.compress_js(
                    [_("pipeline/js/first.js"), _("pipeline/js/second.js")]
                )
            else:
                result = self.compressor.compress_css(
                    [_("pipeline/css/first.css"), _("pipeline/css/second.css")],
                    os.path.join("pipeline", "css", os.path.basename(expected_file)),
                )
        with self.compressor.storage.open(expected_file, "r") as f:
            expected = f.read()
        self.assertEqual(result, expected)

    def test_jsmin(self):
        self._test_compressor(
            "pipeline.compressors.jsmin.JSMinCompressor",
            "js",
            "pipeline/compressors/jsmin.js",
        )

    def test_csshtmljsminify(self):
        self._test_compressor(
            "pipeline.compressors.csshtmljsminify.CssHtmlJsMinifyCompressor",
            "css",
            "pipeline/compressors/csshtmljsminify.css",
        )
        self._test_compressor(
            "pipeline.compressors.csshtmljsminify.CssHtmlJsMinifyCompressor",
            "js",
            "pipeline/compressors/csshtmljsminify.js",
        )

    @skipUnless(settings.HAS_NODE, "requires node")
    def test_uglifyjs(self):
        self._test_compressor(
            "pipeline.compressors.uglifyjs.UglifyJSCompressor",
            "js",
            "pipeline/compressors/uglifyjs.js",
        )

    @skipUnless(settings.HAS_NODE, "requires node")
    def test_terser(self):
        self._test_compressor(
            "pipeline.compressors.terser.TerserCompressor",
            "js",
            "pipeline/compressors/terser.js",
        )

    @skipUnless(settings.HAS_NODE, "requires node")
    def test_yuglify(self):
        self._test_compressor(
            "pipeline.compressors.yuglify.YuglifyCompressor",
            "css",
            "pipeline/compressors/yuglify.css",
        )
        self._test_compressor(
            "pipeline.compressors.yuglify.YuglifyCompressor",
            "js",
            "pipeline/compressors/yuglify.js",
        )

    @skipUnless(settings.HAS_NODE, "requires node")
    def test_cssmin(self):
        self._test_compressor(
            "pipeline.compressors.cssmin.CSSMinCompressor",
            "css",
            "pipeline/compressors/cssmin.css",
        )

    @skipUnless(settings.HAS_NODE, "requires node")
    @skipUnless(settings.HAS_JAVA, "requires java")
    def test_closure(self):
        self._test_compressor(
            "pipeline.compressors.closure.ClosureCompressor",
            "js",
            "pipeline/compressors/closure.js",
        )

    @skipUnless(settings.HAS_NODE, "requires node")
    @skipUnless(settings.HAS_JAVA, "requires java")
    def test_yui_js(self):
        self._test_compressor(
            "pipeline.compressors.yui.YUICompressor",
            "js",
            "pipeline/compressors/yui.js",
        )

    @skipUnless(settings.HAS_NODE, "requires node")
    @skipUnless(settings.HAS_JAVA, "requires java")
    def test_yui_css(self):
        self._test_compressor(
            "pipeline.compressors.yui.YUICompressor",
            "css",
            "pipeline/compressors/yui.css",
        )

    @skipUnless(settings.HAS_CSSTIDY, "requires csstidy")
    def test_csstidy(self):
        self._test_compressor(
            "pipeline.compressors.csstidy.CSSTidyCompressor",
            "css",
            "pipeline/compressors/csstidy.css",
        )



from django.test import TestCase

from pipeline.collector import default_collector
from pipeline.packager import PackageNotFound, Packager
from tests.utils import _


class PackagerTest(TestCase):
    def setUp(self):
        default_collector.collect()

    def test_package_for(self):
        packager = Packager()
        packager.packages["js"] = packager.create_packages(
            {
                "application": {
                    "source_filenames": (_("pipeline/js/application.js"),),
                    "output_filename": "application.js",
                }
            }
        )
        try:
            packager.package_for("js", "application")
        except PackageNotFound:
            self.fail()
        try:
            packager.package_for("js", "broken")
            self.fail()
        except PackageNotFound:
            pass

    def test_templates(self):
        packager = Packager()
        packages = packager.create_packages(
            {
                "templates": {
                    "source_filenames": (_("pipeline/templates/photo/list.jst"),),
                    "output_filename": "templates.js",
                }
            }
        )
        self.assertEqual(
            packages["templates"].templates,
            [_("pipeline/templates/photo/list.jst")],
        )

    def tearDown(self):
        default_collector.clear()



from django.contrib import admin
from django.urls import path
from django.views.generic import TemplateView

urlpatterns = [
    path("", TemplateView.as_view(template_name="index.html"), name="index"),
    path("empty/", TemplateView.as_view(template_name="empty.html"), name="empty"),
    path("admin/", admin.site.urls),
]



from __future__ import annotations

import base64
import os
import posixpath
import re
import subprocess
import warnings
from collections.abc import Iterator, Sequence
from itertools import takewhile

from django.contrib.staticfiles.storage import staticfiles_storage
from django.utils.encoding import force_str, smart_bytes

from pipeline.conf import settings
from pipeline.exceptions import CompressorError
from pipeline.utils import relpath, set_std_streams_blocking, to_class

# Regex matching url(...), url('...'), and url("...") patterns.
#
# Replacements will preserve the quotes and any whitespace contained within
# the pattern, transforming only the filename.
#
# Verbose and documented, to ease future maintenance.
_CSS_URL_REWRITE_PATH_RE_STR = r"""
    (?P<url_prefix>
      url\(                 # The opening `url(`.
      (?P<url_quote>['"]?)  # Optional quote (' or ").
      \s*
    )
    (?P<url_path>.*?)       # The path to capture.
    (?P<url_suffix>
      (?P=url_quote)        # The quote found earlier, if any.
      \s*
      \)                    # The end `)`, completing `url(...)`.
    )
"""


# Regex matching `//@ sourceMappingURL=...` and variants.
#
# This will capture sourceMappingURL and sourceURL keywords, both
# `//@` and `//#` variants, and both `//` and `/* ... */` comment types.
#
# Verbose and documented, to ease future maintenance.
_SOURCEMAP_REWRITE_PATH_RE_STR = r"""
    (?P<sourcemap_prefix>
      /(?:/|(?P<sourcemap_mlcomment>\*))  # Opening comment (`//#`, `//@`,
      [#@]\s+                             # `/*@`, `/*#`).
      source(?:Mapping)?URL=              # The sourcemap indicator.
      \s*
    )
    (?P<sourcemap_path>.*?)               # The path to capture.
    (?P<sourcemap_suffix>
      \s*
      (?(sourcemap_mlcomment)\*/\s*)      # End comment (`*/`)
    )
    $                                     # The line should now end.
"""


# Implementation of the above regexes, for CSS and JavaScript.
CSS_REWRITE_PATH_RE = re.compile(
    f"{_CSS_URL_REWRITE_PATH_RE_STR}|{_SOURCEMAP_REWRITE_PATH_RE_STR}", re.X | re.M
)
JS_REWRITE_PATH_RE = re.compile(_SOURCEMAP_REWRITE_PATH_RE_STR, re.X | re.M)


URL_REPLACER = re.compile(r"""url\(__EMBED__(.+?)(\?\d+)?\)""")
NON_REWRITABLE_URL = re.compile(r"^(#|http:|https:|data:|//)")

DEFAULT_TEMPLATE_FUNC = "template"
TEMPLATE_FUNC = r"""var template = function(str){var fn = new Function('obj', 'var __p=[],print=function(){__p.push.apply(__p,arguments);};with(obj||{}){__p.push(\''+str.replace(/\\/g, '\\\\').replace(/'/g, "\\'").replace(/<%=([\s\S]+?)%>/g,function(match,code){return "',"+code.replace(/\\'/g, "'")+",'";}).replace(/<%([\s\S]+?)%>/g,function(match,code){return "');"+code.replace(/\\'/g, "'").replace(/[\r\n\t]/g,' ')+"__p.push('";}).replace(/\r/g,'\\r').replace(/\n/g,'\\n').replace(/\t/g,'\\t')+"');}return __p.join('');");return fn;};"""  # noqa

MIME_TYPES = {
    ".png": "image/png",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".gif": "image/gif",
    ".tif": "image/tiff",
    ".tiff": "image/tiff",
    ".ttf": "font/truetype",
    ".otf": "font/opentype",
    ".woff": "font/woff",
}
EMBED_EXTS = MIME_TYPES.keys()
FONT_EXTS = [".ttf", ".otf", ".woff"]


class Compressor:
    asset_contents = {}

    def __init__(self, storage=None, verbose=False):
        if storage is None:
            storage = staticfiles_storage
        self.storage = storage
        self.verbose = verbose

    @property
    def js_compressor(self):
        return to_class(settings.JS_COMPRESSOR)

    @property
    def css_compressor(self):
        return to_class(settings.CSS_COMPRESSOR)

    def compress_js(
        self,
        paths: Sequence[str],
        templates: Sequence[str] | None = None,
        *,
        output_filename: str | None = None,
        **kwargs,
    ) -> str:
        """Concatenate and compress JS files"""
        # Note how a semicolon is added between the two files to make sure that
        # their behavior is not changed. '(expression1)\n(expression2)' calls
        # `expression1` with `expression2` as an argument! Superfluous
        # semicolons are valid in JavaScript and will be removed by the
        # minifier.
        js = self.concatenate(
            paths,
            file_sep=";",
            output_filename=output_filename,
            rewrite_path_re=JS_REWRITE_PATH_RE,
        )

        if templates:
            js = js + self.compile_templates(templates)

        if not settings.DISABLE_WRAPPER:
            js = settings.JS_WRAPPER % js

        compressor = self.js_compressor
        if compressor:
            js = getattr(compressor(verbose=self.verbose), "compress_js")(js)

        return js

    def compress_css(self, paths, output_filename, variant=None, **kwargs):
        """Concatenate and compress CSS files"""
        css = self.concatenate(
            paths,
            file_sep="",
            rewrite_path_re=CSS_REWRITE_PATH_RE,
            output_filename=output_filename,
            variant=variant,
        )
        compressor = self.css_compressor
        if compressor:
            css = getattr(compressor(verbose=self.verbose), "compress_css")(css)
        if not variant:
            return css
        elif variant == "datauri":
            return self.with_data_uri(css)
        else:
            raise CompressorError(f'"{variant}" is not a valid variant')

    def compile_templates(self, paths):
        compiled = []
        if not paths:
            return ""
        namespace = settings.TEMPLATE_NAMESPACE
        base_path = self.base_path(paths)
        for path in paths:
            contents = self.read_text(path)
            contents = re.sub("\r?\n", "\\\\n", contents)
            contents = re.sub("'", "\\'", contents)
            name = self.template_name(path, base_path)
            compiled.append(
                "{}['{}'] = {}('{}');\n".format(
                    namespace, name, settings.TEMPLATE_FUNC, contents
                )
            )
        if settings.TEMPLATE_FUNC == DEFAULT_TEMPLATE_FUNC:
            compiler = TEMPLATE_FUNC
        else:
            compiler = ""
        return "\n".join(
            [
                "{namespace} = {namespace} || {{}};".format(namespace=namespace),
                compiler,
                "".join(compiled),
            ]
        )

    def base_path(self, paths):
        def names_equal(name):
            return all(n == name[0] for n in name[1:])

        directory_levels = zip(*[p.split(os.sep) for p in paths])
        return os.sep.join(x[0] for x in takewhile(names_equal, directory_levels))

    def template_name(self, path, base):
        """Find out the name of a JS template"""
        if not base:
            path = os.path.basename(path)
        if path == base:
            base = os.path.dirname(path)
        name = re.sub(
            r"^{}[\/\\]?(.*){}$".format(
                re.escape(base), re.escape(settings.TEMPLATE_EXT)
            ),
            r"\1",
            path,
        )
        return re.sub(r"[\/\\]", settings.TEMPLATE_SEPARATOR, name)

    def concatenate_and_rewrite(self, paths, output_filename, variant=None):
        """Concatenate together files and rewrite urls"""
        warnings.warn(
            "Compressor.concatenate_and_rewrite() is deprecated. Please "
            "call concatenate() instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self.concatenate(
            paths=paths,
            file_sep="",
            rewrite_path_re=CSS_REWRITE_PATH_RE,
            output_filename=output_filename,
            variant=variant,
        )

    def concatenate(
        self,
        paths: Sequence[str],
        *,
        file_sep: str | None = None,
        output_filename: str | None = None,
        rewrite_path_re: re.Pattern | None = None,
        variant: str | None = None,
    ) -> str:
        """Concatenate together a list of files.

        The caller can specify a delimiter between files and any regexes
        used to normalize relative paths. Path normalization is important for
        ensuring that local resources or sourcemaps can be updated in time
        for Django's static media post-processing phase.
        """

        def _reconstruct(
            m: re.Match,
            source_path: str,
        ) -> str:
            groups = m.groupdict()
            asset_path: str | None = None
            prefix = ""
            suffix = ""

            for prefix in ("sourcemap", "url"):
                asset_path = groups.get(f"{prefix}_path")

                if asset_path is not None:
                    asset_path = asset_path.strip()
                    prefix, suffix = m.group(f"{prefix}_prefix", f"{prefix}_suffix")
                    break

            if asset_path is None:
                # This is empty. Return the whole match as-is.
                return m.group()

            if asset_path and not NON_REWRITABLE_URL.match(asset_path):
                asset_path = self.construct_asset_path(
                    asset_path=asset_path,
                    source_path=source_path,
                    output_filename=output_filename,
                    variant=variant,
                )

            return f"{prefix}{asset_path}{suffix}"

        def _iter_files() -> Iterator[str]:
            if not output_filename or not rewrite_path_re:
                # This is legacy call, which does not support sourcemap-aware
                # asset rewriting. Pipeline itself won't invoke this outside
                # of tests, but it maybe important for third-parties who
                # are specializing these classes.
                warnings.warn(
                    "Compressor.concatenate() was called without passing "
                    "rewrite_path_re_= or output_filename=. If you are "
                    "specializing Compressor, please update your call "
                    "to remain compatible with future changes.",
                    DeprecationWarning,
                    stacklevel=3,
                )

                return (self.read_text(path) for path in paths)

            # Now that we can attempt the modern support for concatenating
            # files, handling rewriting of relative assets in the process.
            return (
                rewrite_path_re.sub(
                    lambda m: _reconstruct(m, path), self.read_text(path)
                )
                for path in paths
            )

        if file_sep is None:
            warnings.warn(
                "Compressor.concatenate() was called without passing "
                "file_sep=. If you are specializing Compressor, please "
                "update your call to remain compatible with future changes. "
                "Defaulting to JavaScript behavior for "
                "backwards-compatibility.",
                DeprecationWarning,
                stacklevel=2,
            )
            file_sep = ";"

        return f"\n{file_sep}".join(_iter_files())

    def construct_asset_path(
        self, asset_path, source_path, output_filename, variant=None
    ):
        """Return a rewritten asset URL for a stylesheet or JavaScript file."""
        public_path = self.absolute_path(
            asset_path,
            os.path.dirname(source_path).replace("\\", "/"),
        )
        if self.embeddable(public_path, variant):
            return f"__EMBED__{public_path}"
        if not posixpath.isabs(asset_path):
            asset_path = self.relative_path(public_path, output_filename)
        return asset_path

    def embeddable(self, path, variant):
        """Is the asset embeddable ?"""
        name, ext = os.path.splitext(path)
        font = ext in FONT_EXTS
        if not variant:
            return False
        if not (
            re.search(settings.EMBED_PATH, path.replace("\\", "/"))
            and self.storage.exists(path)
        ):
            return False
        if ext not in EMBED_EXTS:
            return False
        if not (
            font or len(self.encoded_content(path)) < settings.EMBED_MAX_IMAGE_SIZE
        ):
            return False
        return True

    def with_data_uri(self, css):
        def datauri(match):
            path = match.group(1)
            mime_type = self.mime_type(path)
            data = self.encoded_content(path)
            return f'url("data:{mime_type};charset=utf-8;base64,{data}")'

        return URL_REPLACER.sub(datauri, css)

    def encoded_content(self, path):
        """Return the base64 encoded contents"""
        if path in self.__class__.asset_contents:
            return self.__class__.asset_contents[path]
        data = self.read_bytes(path)
        self.__class__.asset_contents[path] = force_str(base64.b64encode(data))
        return self.__class__.asset_contents[path]

    def mime_type(self, path):
        """Get mime-type from filename"""
        name, ext = os.path.splitext(path)
        return MIME_TYPES[ext]

    def absolute_path(self, path, start):
        """
        Return the absolute public path for an asset,
        given the path of the stylesheet that contains it.
        """
        if posixpath.isabs(path):
            path = posixpath.join(staticfiles_storage.location, path)
        else:
            path = posixpath.join(start, path)
        return posixpath.normpath(path)

    def relative_path(self, absolute_path, output_filename):
        """Rewrite paths relative to the output stylesheet path"""
        absolute_path = posixpath.join(settings.PIPELINE_ROOT, absolute_path)
        output_path = posixpath.join(
            settings.PIPELINE_ROOT, posixpath.dirname(output_filename)
        )
        return relpath(absolute_path, output_path)

    def read_bytes(self, path):
        """Read file content in binary mode"""
        file = staticfiles_storage.open(path)
        content = file.read()
        file.close()
        return content

    def read_text(self, path):
        content = self.read_bytes(path)
        return force_str(content)


class CompressorBase:
    def __init__(self, verbose):
        self.verbose = verbose

    def filter_css(self, css):
        raise NotImplementedError

    def filter_js(self, js):
        raise NotImplementedError


class SubProcessCompressor(CompressorBase):
    def execute_command(self, command, content):
        argument_list = []
        for flattening_arg in command:
            if isinstance(flattening_arg, (str,)):
                argument_list.append(flattening_arg)
            else:
                argument_list.extend(flattening_arg)

        pipe = subprocess.Popen(
            argument_list,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if content:
            content = smart_bytes(content)
        stdout, stderr = pipe.communicate(content)
        set_std_streams_blocking()
        if stderr.strip() and pipe.returncode != 0:
            raise CompressorError(force_str(stderr))
        elif self.verbose:
            print(force_str(stderr))
        return force_str(stdout)


class NoopCompressor(CompressorBase):
    def compress_js(self, js):
        return js

    def compress_css(self, css):
        return css
